/*
 * SeqQPCET.cpp
 *
 *  Created on: Jun 3, 2019
 *      Author: neanderthal
 */

#include<cmath>
#include <unistd.h>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <complex>
#include <omp.h>
#include<sstream>
#include<string>

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/opencv.hpp>
#include <opencv2/imgproc.hpp>

#include <sys/time.h>
#include <vector>
#include <fstream>

using namespace std;
using namespace cv;

vector<vector<int> > blue;
vector<vector<int> > green;
vector<vector<int> > red;
String imageFileName;

const int SIZE = 256;
#define MAX 20

const int S = 4;
const int M = SIZE / 2;
const int Theta1Col = (2 * (M - 1) + 1) * S;

unsigned short numOfThreads = 16;

vector<vector<double> > fR, fG, fB;
std::complex<double> SR[MAX + 1][MAX + 1];
std::complex<double> SG[MAX + 1][MAX + 1];
std::complex<double> SB[MAX + 1][MAX + 1];
std::complex<double> H1R[MAX + 1][MAX + 1];
std::complex<double> H1G[MAX + 1][MAX + 1];
std::complex<double> H1B[MAX + 1][MAX + 1];

vector<vector<double> > h_1, h_2, h_3;

vector<double> interker(vector<double> x) {
	vector<double> hx(SIZE);
	double Absx[SIZE];
	for (unsigned i = 0; i < x.size(); i++) {
		Absx[i] = abs(x[i]);
	}

	for (unsigned i = 0; i < x.size(); i++) {
		if (Absx[i] <= 1)
			hx[i] = (3.0 / 2.0) * pow(Absx[i], 3)
					- (5.0 / 2.0) * pow(Absx[i], 2) + 1;
		else if (Absx[i] <= 2)
			hx[i] = (-1.0 / 2) * pow(Absx[i], 3) + (5.0 / 2.0) * pow(Absx[i], 2)
					- 4 * Absx[i] + 2;
		else
			hx[i] = 0;
	}
	return hx;
}

void paralell_4point_symmetry_QuaternionMoment() {
	struct timeval startAccesstime, endAccesstime;

	int Ki[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ki[i] = (2 * i + 1) * S;
	}

	double Ri[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ri[i] = (2 * i + 1) / (2.0 * M);
	}

	double deltaR = 1.0 / M;
	double U[M + 1];
	U[0] = Ri[0] - (deltaR / 2.0);

	for (int i = 1; i <= M; i++) {
		U[i] = Ri[i - 1] + deltaR / 2.0;
	}

	vector<vector<double> > Theta1;
	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		Theta1.push_back(x);
	}

	double PI = 3.1415926535897;

	for (int i = 1; i <= M; i++) {
		for (int n = 0; n < Ki[i - 1]; n++) {
			Theta1[i - 1][n] = (n + 0.5) * (2 * PI / Ki[i - 1]);
		}
	}

	//the upper and lower limit of the sectors angles
	double deltaTheta;

	vector<vector<double> > V;
	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		V.push_back(x);
	}

	for (int i = 0; i < M; i++) {
		deltaTheta = 2 * PI / Ki[i];
		V[i][0] = Theta1[i][0] - (deltaTheta / 2.0);
		for (int n = 1; n < Ki[i] + 1; n++) {
			V[i][n] = Theta1[i][n - 1] + (deltaTheta / 2.0);
		}
	}

	// Image interolation
	double r[M];
	for (int i = 0; i < M; i++)
		r[i] = Ri[i];

//compute xi,yi
	int s[SIZE];
	for (int i = 0; i < SIZE; i++)
		s[i] = i + 1;

	double xi[SIZE];
	for (int i = 0; i < SIZE; i++)
		xi[i] = (2.0 * s[i] - SIZE - 1) / SIZE;

	double yi[SIZE];
	for (int i = 0; i < SIZE; i++)
		yi[i] = (2.0 * s[i] - SIZE - 1) / SIZE;

	double del = 2.0 / SIZE;   //% the pixel width
	const int N = SIZE;

	gettimeofday(&startAccesstime, NULL);

	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		fR.push_back(x);
		fB.push_back(x);
		fG.push_back(x);
	}

#pragma omp parallel for schedule(dynamic,1) num_threads(numOfThreads)
	for (int i = 1; i <= M; i++) {
		for (int j = 1; j <= Ki[i - 1]; j++) {

			int k1 = ceil(r[i - 1] * cos(Theta1[i - 1][j - 1]) / del) + N / 2;
			int k2 = ceil(r[i - 1] * sin(Theta1[i - 1][j - 1]) / del) + N / 2;

			int lSize = (k1 + 2) - (k1 - 1) + 1;
			int mSize = (k2 + 2) - (k2 - 1) + 1;

			std::vector<int> l(lSize);
			std::vector<int> m(mSize);
			std::vector<int> inx(lSize);
			std::vector<int> iny(mSize);

			for (int i = 0, count = k1 - 1; i < lSize; i++) {
				l[i] = count++;
				inx[i] = l[i];
			}

			for (int i = 0, count = k2 - 1; i < mSize; i++) {
				m[i] = count++;
				iny[i] = m[i];
			}

			// construct image matrix
			for (int i = 0; i < lSize; i++) {
				if (l[i] > N) {
					inx[i] = l[i] % N;
					l[i] = N - (l[i] % N) + 1;
					continue;
				}
				if (l[i] < 1) {
					inx[i] = N - abs(l[i]);
					l[i] = N - ((l[i] + N) - 1);
				}

			}

			for (int i = 0; i < mSize; i++) {
				if (m[i] > N) {
					iny[i] = m[i] % N;
					m[i] = N - (m[i] % N) + 1;
					continue;
				}
				if (m[i] < 1) {
					iny[i] = N - abs(m[i]);
					m[i] = N - ((m[i] + N) - 1);
				}

			}
			//%construct the image 4*4 matrix
			vector<vector<double> > Imatrix1(lSize, vector<double>(mSize));
			vector<vector<double> > Imatrix2(lSize, vector<double>(mSize));
			vector<vector<double> > Imatrix3(lSize, vector<double>(mSize));

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix1[i][j] = red[l[i] - 1][m[j] - 1];
				}
			}

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix2[i][j] = green[l[i] - 1][m[j] - 1];
				}
			}

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix3[i][j] = blue[l[i] - 1][m[j] - 1];
				}
			}
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//% construct kernel
			std::vector<double> xl(lSize);

			for (int i = 0; i < lSize; i++)
				xl[i] = xi[inx[i] - 1];

			std::vector<double> ym(mSize);

			for (int i = 0; i < mSize; i++)
				ym[i] = yi[iny[i] - 1];

			// % compute the kernel function using 'interker' function
			std::vector<double> tempXl(lSize);

			for (int ww = 0; ww < lSize; ww++) {
				tempXl[ww] = (r[i - 1] * cos(Theta1[i - 1][j - 1]) - xl[ww])
						/ del;
			}

			std::vector<double> h1(lSize);

			vector<double> hx = interker(tempXl);  //% for x
			for (int i = 0; i < lSize; i++)
				h1[i] = hx[i];

			std::vector<double> tempYm(mSize);

			for (int ww = 0; ww < mSize; ww++)
				tempYm[ww] = (r[i - 1] * sin(Theta1[i - 1][j - 1]) - ym[ww])
						/ del;

			std::vector<double> h2(mSize);

			hx = interker(tempYm);  //% for x
			for (int i = 0; i < lSize; i++)
				h2[i] = hx[i];

			vector<vector<double> > hh(lSize, vector<double>(mSize));

			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++)
					hh[i][j] = h1[i] * h2[j];

			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//% 2D convolution between the kernel and the image matrix
			vector<vector<double> > P1(lSize, vector<double>(mSize));
			vector<vector<double> > P2(lSize, vector<double>(mSize));
			vector<vector<double> > P3(lSize, vector<double>(mSize));

			double sum1 = 0, sum2 = 0, sum3 = 0;
			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P1[i][j] = Imatrix1[i][j] * hh[i][j];
					sum1 += P1[i][j];
				}

			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P2[i][j] = Imatrix2[i][j] * hh[i][j];
					sum2 += P2[i][j];
				}

			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P3[i][j] = Imatrix3[i][j] * hh[i][j];
					sum3 += P3[i][j];
				}
			fR[i - 1][j - 1] = sum1;
			fG[i - 1][j - 1] = sum2;
			fB[i - 1][j - 1] = sum3;

			//clearing area
			P1.clear();
			P1.clear();
			P3.clear();
			hh.clear();
			h2.clear();
			tempYm.clear();
			h1.clear();
			tempXl.clear();
			ym.clear();
			xl.clear();
			Imatrix3.clear();
			Imatrix2.clear();
			Imatrix1.clear();
			iny.clear();
			inx.clear();
			m.clear();
			l.clear();

		}
	}
	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "\nImage Interpolation time = " << totTime << " seconds." << endl;

	gettimeofday(&startAccesstime, NULL);
	int pmax = MAX;
	int qmax = MAX;

	std::complex<double> MM(0, 1);
	std::complex<double> I3(0, 1);
	std::complex<double> i1neg(0, -1);

	vector<vector<std::complex<double> > > Ip(MAX + 1,
			vector<std::complex<double> >(M));

#pragma omp parallel for private(I3) num_threads(numOfThreads)
	for (int p = 0; p <= pmax; p++) {
		for (int i = 1; i <= M; i++) {
			if (p == 0)
				I3 = 0.5 * (pow(U[i], 2.0) - pow(U[i - 1], 2.0));
			else
				I3 = (exp(p * 2.0 * i1neg * PI * pow(U[i], 2.0))
						- exp(p * 2.0 * i1neg * PI * pow(U[i - 1], 2.0)))
						/ (p * 4.0 * PI * i1neg);
			Ip[p][i - 1] = I3;
		}
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "I_p time = " << totTime << " seconds." << endl;

	typedef vector<std::complex<double> > v1d;
	typedef vector<v1d> v2d;
	typedef vector<v2d> v3d;

	std::complex<double> I(0, 0);
	std::complex<double> i1pos(0, 1);

	v3d Iq;
	for (int i = 0; i < M; i++) {
		v2d x(MAX + 1, v1d(S * (2 * i + 1)));
		Iq.push_back(x);
	}

	gettimeofday(&startAccesstime, NULL);

	//% %%%%%%%%%%%%%%%%%%%%%%I(q)%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#pragma omp parallel for private(I) num_threads(numOfThreads)
	for (int q = 0; q <= qmax; q++) {
		for (int i = 1; i <= M; i++) {
			for (int j = 1; j <= Ki[i - 1]; j++) {
				if (q == 0)
					I = V[i - 1][j] - V[i - 1][j - 1];
				else
					I = (i1pos / (double) q)
							* (exp(-1.0 * q * i1pos * V[i - 1][j])
									- exp(-1.0 * q * i1pos * V[i - 1][j - 1]));
				Iq[i - 1][q][j - 1] = I;
			}
		}
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "I_q time = " << totTime << " seconds." << endl;

	//%%%%%%%%%%%%%%%%%%%%%%%%%QPCET _  Moment Computed.................
	for (int i = 0; i < MAX + 1; i++) {
		for (int j = 0; j < MAX + 1; j++) {
			SR[i][j] = SG[i][j] = SB[i][j] = 0;
		}
	}

	gettimeofday(&startAccesstime, NULL);
	omp_set_num_threads(numOfThreads);

	int loopSize = MAX + 1;
	int size = loopSize * loopSize / numOfThreads;
	double pii = 7.0 / 22.0;

#pragma omp parallel num_threads(numOfThreads)
	//for(int threadID = 0, rr =0 ; threadID < numOfThreads; threadID++)
	{
		int threadID = omp_get_thread_num();
		std::complex<double> S1 = 0, S2 = 0, S3 = 0, S4 = 0, S5 = 0, S6 = 0;
		std::complex<double> YR = 0, YG = 0, YB = 0;

		int local_start = threadID * size;
		int local_end = (threadID + 1) * size - 1;
		if (threadID == numOfThreads - 1)
			local_end = loopSize * loopSize - 1;

		int p, q;
		int z1, z2, z3;

		for (int k = local_start; k <= local_end; k++) {
			p = k / (loopSize);
			q = k % loopSize;
			S1 = S2 = S3 = S4 = S5 = S6 = 0;
			for (int i = 1; i <= M; i++) {
				for (int j = 1; j <= 2 * i - 1; j++) {
					z1 = 4 * i - 1 - j;
					z2 = 4 * i - 2 + j;
					z3 = 8 * i - 3 - j;

					YR = fR[i - 1][j - 1] * Iq[i - 1][q][j - 1]
							+ fR[i - 1][z1 - 1] * Iq[i - 1][q][z1 - 1]
							+ fR[i - 1][z2 - 1] * Iq[i - 1][q][z2 - 1]
							+ fR[i - 1][z3 - 1] * Iq[i - 1][q][z3 - 1];
					YG = fG[i - 1][j - 1] * Iq[i - 1][q][j - 1]
							+ fG[i - 1][z1 - 1] * Iq[i - 1][q][z1 - 1]
							+ fG[i - 1][z2 - 1] * Iq[i - 1][q][z2 - 1]
							+ fG[i - 1][z3 - 1] * Iq[i - 1][q][z3 - 1];
					YB = fB[i - 1][j - 1] * Iq[i - 1][q][j - 1]
							+ fB[i - 1][z1 - 1] * Iq[i - 1][q][z1 - 1]
							+ fB[i - 1][z2 - 1] * Iq[i - 1][q][z2 - 1]
							+ fB[i - 1][z3 - 1] * Iq[i - 1][q][z3 - 1];

					S1 = S1 + (YR) * Ip[p][i - 1];
					S2 = S2 + (YG) * Ip[p][i - 1];
					S3 = S3 + (YB) * Ip[p][i - 1];

					S4 = S4 + (YR) * conj(Ip[p][i - 1]);
					S5 = S5 + (YG) * conj(Ip[p][i - 1]);
					S6 = S6 + (YB) * conj(Ip[p][i - 1]);
				}
			}

			SR[p][q] = pii * S1;
			SG[p][q] = pii * S2;
			SB[p][q] = pii * S3;
			H1R[p][q] = pii * S4;
			H1G[p][q] = pii * S5;
			H1B[p][q] = pii * S6;
		}
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Parallel quater. 4-point symmetry moment computation for order "
			<< MAX << ": running time = " << totTime << " seconds, using "
			<< numOfThreads << " threads." << endl;
}

void paralell_image_reconstruction() {

	double x[SIZE];

	for (int i = 1; i <= SIZE; i++) {
		x[i - 1] = ((2 * i) - SIZE - 1) / (double) SIZE;
	}

	vector<vector<double> > Theta1(SIZE, vector<double>(SIZE));
	vector<vector<double> > r(SIZE, vector<double>(SIZE));

	for (int i = 0; i < SIZE; i++)
		for (int j = 0; j < SIZE; j++) {
			r[i][j] = sqrt(pow(x[i], 2) + pow(x[j], 2));
			Theta1[i][j] = std::atan2(x[j], x[i]);
		}
	struct timeval startAccesstime, endAccesstime;

	gettimeofday(&startAccesstime, NULL);
	std::complex<double> sum3 = 0.0, sum2 = 0.0, sum1 = 0.0;
	std::complex<double> s1 = 0.0, s2 = 0.0, s3 = 0.0;

	for (int i = 0; i < SIZE; i++) {
		vector<double> x(SIZE);
		std::fill(x.begin(), x.end(), 0);
		h_1.push_back(x);
		h_2.push_back(x);
		h_3.push_back(x);
	}

	int pmax, qmax;
	pmax = qmax = MAX;
	omp_set_num_threads(numOfThreads);

	std::complex<double> ii(0, 1);
	const double PI = 3.1415926535897;

	int threadLoad = SIZE / numOfThreads;
#pragma omp parallel private(sum1, sum2,sum3,s1,s2,s3) num_threads(numOfThreads)
//for(int threadID = 0 ; threadID < numOfThreads; threadID++){
	{
		int threadID = omp_get_thread_num();
		int start = threadID * threadLoad + 1;
		int end = (threadID + 1) * threadLoad;
		for (int i = start; i <= end; i++) {
			for (int j = 1; j <= SIZE; j++) {
				std::complex<double> R1, R2;

				for (int p = 1; p <= pmax; p++) {
					R1 = exp(p * 2.0 * ii * PI * pow(r[i - 1][j - 1], 2.0));
					R2 = exp(p * Theta1[i - 1][j - 1] * ii);

					s1 = s1 + (SR[p][0]) * (R1) + conj(SR[p][0]) * conj(R1)
							+ (SR[0][p]) * (R2) + conj(SR[0][p] * (R2));
					s2 = s2 + (SG[p][0]) * (R1) + conj(SG[p][0]) * conj(R1)
							+ (SG[0][p]) * (R2) + conj(SG[0][p] * (R2));
					s3 = s3 + (SB[p][0]) * (R1) + conj(SB[p][0]) * conj(R1)
							+ (SB[0][p]) * (R2) + conj(SB[0][p] * (R2));

					for (int q = 1; q <= qmax; q++) {
						R2 = exp(q * Theta1[i - 1][j - 1] * ii);
						sum1 = sum1 + SR[p][q] * R1 * R2
								+ conj(SR[p][q] * R1 * R2)
								+ H1R[p][q] * conj(R1) * (R2)
								+ conj(H1R[p][q]) * R1 * conj(R2);
						sum2 = sum2 + SG[p][q] * R1 * R2
								+ conj(SG[p][q] * R1 * R2)
								+ H1G[p][q] * conj(R1) * (R2)
								+ conj(H1G[p][q]) * R1 * conj(R2);
						sum3 = sum3 + SB[p][q] * R1 * R2
								+ conj(SB[p][q] * R1 * R2)
								+ H1B[p][q] * conj(R1) * (R2)
								+ conj(H1B[p][q]) * R1 * conj(R2);
					}
				}

				h_3[(i - 1)][(j - 1)] = (sum3 + s3 + SB[0][0]).real();
				h_2[(i - 1)][(j - 1)] = (sum2 + s2 + SG[0][0]).real();
				h_1[(i - 1)][(j - 1)] = (sum1 + s1 + SR[0][0]).real();
				sum1 = sum2 = sum3 = s1 = s2 = s3 = 0;
			}
		}
	}

	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Parallel quater. image reconstruct time = " << totTime
			<< " seconds." << endl;
}

void errorComputing() {

	double x[SIZE];
	for (int i = 1; i <= SIZE; i++) {
		x[i - 1] = (2 * i - SIZE - 1) / (double) SIZE;
	}
	vector<vector<double> > r(SIZE, vector<double>(SIZE));

	for (int i = 0; i < SIZE; i++)
		for (int j = 0; j < SIZE; j++) {
			r[i][j] = sqrt(pow(x[i], 2) + pow(x[j], 2));
		}
	//%%%%%%%%%%NIRE Error%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	double sun1 = 0.0, sun11 = 0.0, sun22 = 0.0, sun222 = 0.0, sun2 = 0.0,
			sun3 = 0.0;
	for (int i = 1; i <= SIZE; i++) {
		for (int j = 1; j <= SIZE; j++) {
			if (r[i - 1][j - 1] <= 1) {

				sun2 = sun2 + pow(blue[i - 1][j - 1], 2.0);
				sun3 = sun3 + pow(h_3[i - 1][j - 1] - blue[i - 1][j - 1], 2.0);
				// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				sun1 = sun1 + pow(red[i - 1][j - 1], 2.0);
				sun11 = sun11 + pow(h_1[i - 1][j - 1] - red[i - 1][j - 1], 2.0);
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				sun22 = sun22 + pow(green[i - 1][j - 1], 2.0);
				sun222 = sun222
						+ pow(h_2[i - 1][j - 1] - green[i - 1][j - 1], 2.0);
			} else {
				h_3[(i - 1)][(j - 1)] = h_2[(i - 1)][(j - 1)] = h_1[(i - 1)][(j
						- 1)] = 255;
			}

		}
	}
	cout << endl;
	double dif3 = (sun3 / sun2);
	double dif2 = (sun222 / sun22);
	double dif1 = (sun11 / sun1);
	double dif = (dif2 + dif3 + dif1) / 3.0;
	cout << "Difference between original and reconstructed image = " << dif
			<< endl;
}

bool saveReconstructedImage(bool showImage = false) {
	cv::Mat mat_red(h_1.size(), h_1.at(0).size(), CV_64FC1);
	for (int i = 0; i < mat_red.rows; ++i)
		for (int j = 0; j < mat_red.cols; ++j)
			mat_red.at<double>(i, j) = h_1.at(i).at(j);

	cv::Mat mat_green(h_2.size(), h_2.at(0).size(), CV_64FC1);
	for (int i = 0; i < mat_green.rows; ++i)
		for (int j = 0; j < mat_green.cols; ++j)
			mat_green.at<double>(i, j) = h_2.at(i).at(j);

	cv::Mat mat_blue(h_3.size(), h_3.at(0).size(), CV_64FC1);
	for (int i = 0; i < mat_blue.rows; ++i)
		for (int j = 0; j < mat_blue.cols; ++j)
			mat_blue.at<double>(i, j) = h_3.at(i).at(j);

	Mat temp;
	vector<Mat> channels;

	channels.push_back(mat_blue);
	channels.push_back(mat_green);
	channels.push_back(mat_red);

	merge(channels, temp);
	Mat final_reconstructed_image;
	temp.convertTo(final_reconstructed_image, CV_8UC1);
	String imName = "reconstructed-" + std::to_string(MAX) + "-"
			+ imageFileName;
	imwrite(imName, final_reconstructed_image);
	if (showImage) {
		imshow(">>Reconst. Image", final_reconstructed_image);
		waitKey(0);
	}
	cout << "The reconstructed image is saved on disk with file name: "
			<< imName << endl;
	return true;
}

bool readImage(const String imageName) {

	ifstream image_file(imageName.c_str());

	if (!image_file.good()) {
		cerr << "Error: the image with name \"" << imageName
				<< "\" doesn't exist." << endl;
		return false;
	}
	cout << "Reading image: " << imageName << "......";
	Mat image = cv::imread(imageName);
	if (image.channels() != 3) {
		cerr << "Error: only images with three channels are supported" << endl;
		return false;
	}
	int tempRow = -1;
	int tempCol = -1;
	if (image.rows != SIZE || image.cols != SIZE) {
		tempRow = image.rows;
		tempCol = image.cols;
		resize(image, image, Size(SIZE, SIZE));
	}
	Mat bgr[3];   //destination array
	split(image, bgr);   //split source

	//Note: OpenCV uses BGR color order
	imwrite("blue.png", bgr[0]);   //blue channel
	imwrite("green.png", bgr[1]);   //green channel
	imwrite("red.png", bgr[2]);   //red channel

	vector<vector<int> > temp_blue(bgr[0].rows, vector<int>(bgr[0].cols));
	vector<vector<int> > temp_green(bgr[1].rows, vector<int>(bgr[1].cols));
	vector<vector<int> > temp_red(bgr[2].rows, vector<int>(bgr[2].cols));

	for (int i = 0; i < bgr[0].rows; i++)
		for (int j = 0; j < bgr[0].cols; j++) {
			if ((int) bgr[0].at<char>(i, j) >= 0) {
				temp_blue[i][j] = (int) bgr[0].at<char>(i, j);
			} else {
				temp_blue[i][j] = 256 + (int) bgr[0].at<char>(i, j);
			}
			if ((int) bgr[1].at<char>(i, j) >= 0) {
				temp_green[i][j] = (int) bgr[1].at<char>(i, j);
			} else {
				temp_green[i][j] = 256 + (int) bgr[1].at<char>(i, j);
			}

			if ((int) bgr[2].at<char>(i, j) >= 0) {
				temp_red[i][j] = (int) bgr[2].at<char>(i, j);
			} else {
				temp_red[i][j] = 256 + (int) bgr[2].at<char>(i, j);
			}
		}

	blue = temp_blue;
	green = temp_green;
	red = temp_red;

	cout << "Done!" << endl;
	if (tempCol != -1)
		cout << "The input image size is:" << tempRow << "x" << tempCol
				<< ". The image is resized to this dimensions: " << SIZE << "x"
				<< SIZE << "." << endl;
	return true;
}

static void show_usage(std::string name) {
	std::cerr << "Usage: " << name << " <option(s)>: \n"
			<< "\t--image\tthe input file image\n"
			<< "\t--threads\tSpecify the used number of threads." << std::endl;
}

bool set_inputs(int argc, const char * argv[]) {
	if (argc < 3) {
		show_usage(argv[0]);
		return false;
	}

	for (int i = 1; i < argc; ++i) {
		std::string arg = argv[i];
		if (arg == "--image") {
			imageFileName = argv[i + 1];
			i++;
			if (!readImage(imageFileName)) {
				return false;
			}
		} else if (arg == "--threads") {
			numOfThreads = atoi(argv[i + 1]);
			i++;
			if (numOfThreads < 2) {
				std::cerr << "The number of threads should be more than 1."
						<< endl;
				return false;
			}
			if (numOfThreads > (MAX + 1) * (MAX + 1)) {
				numOfThreads = (MAX + 1) * (MAX + 1);
				std::cerr << "The number of threads is adjusted to "
						<< numOfThreads
						<< ". The maximum allowed number of thread is (moment order +1)^2 = ("
						<< MAX << " + 1)^2." << endl;
			}
			cout << "The number of used threads is set to: " << numOfThreads
					<< " threads." << endl;
		} else {
			std::cerr << argv[i] << " is an invalid argument." << endl;
			return false;
		}
	}
	return true;
}

int main(int argc, const char * argv[]) {
	if (!set_inputs(argc, argv))
		return 0;

	omp_set_dynamic(0);
	omp_set_num_threads(numOfThreads);

#pragma omp parallel
	{
		int ID = omp_get_thread_num();
	}
	paralell_4point_symmetry_QuaternionMoment();
	paralell_image_reconstruction();
	errorComputing();
	saveReconstructedImage();
	cout << "All tasks are done!" << endl;
}

