/**
 * Copyright 1993-2012 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 */
#include <stdio.h>
#include<cmath>
#include <unistd.h>
#include <numeric>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <complex>
#include <omp.h>
#include<sstream>
#include<string>

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/opencv.hpp>
#include <opencv2/imgproc.hpp>

#include <sys/time.h>
#include <vector>
#include <fstream>
#include <thrust/complex.h>

using namespace std;
using namespace cv;

#include <cuda.h>

int numOfThreadsCPU = 16;

int noOfBlocks = 10;
int noOfThreads = 512;
int totNoOfThreads = 0;
char * fileName;

const int SIZE = 256;
#define MAX 20

const int M = SIZE / 2;
const int S = 4;

vector<vector<int> > blue;
vector<vector<int> > green;
vector<vector<int> > red;
String imageFileName;

int numOfDevices = 1;

//image independent data
thrust::complex<double> * d_Ip;
thrust::complex<double> * d_Iq;

typedef vector<std::complex<double> > v1d;
typedef vector<v1d> v2d;
typedef vector<v2d> v3d;
v3d Iq;

vector<vector<std::complex<double> > > Ip(MAX + 1,
		vector<std::complex<double> >(M));
vector<vector<double> > Theta1;

//image dependent
vector<vector<double> > fR, fG, fB;

//global moment data
thrust::complex<double>* h_SB;
thrust::complex<double>* h_SG;
thrust::complex<double>* h_SR;
thrust::complex<double>* h_H1B;
thrust::complex<double>* h_H1G;
thrust::complex<double>* h_H1R;

double * h_1;
double * h_2;
double * h_3;

inline void gpuAssert(cudaError_t code, const char *file, int line, bool abort =
		true) {
	if (code != cudaSuccess) {
		fprintf(stderr, "GPUassert: %s %s %d\n", cudaGetErrorString(code), file,
				line);
		if (abort)
			exit(code);
	}
}
#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }

__global__ void QuaternionMomentKernel(thrust::complex<double> *Ip,
		thrust::complex<double> *Iq, double *fR, double *fB, double *fG,
		thrust::complex<double> *SR, thrust::complex<double> *SB,
		thrust::complex<double> *SG, thrust::complex<double> *H1R,
		thrust::complex<double> *H1B, thrust::complex<double> *H1G, int m,
		int max, int numOfThreads) {

	double pii = 7.0 / 22.0;

	int loopSize = max + 1;
	int p, q;
	int z1, z2, z3, i_index, q_index, p_index;

	thrust::complex<double> S1 = 0, S2 = 0, S3 = 0, S4 = 0, S5 = 0, S6 = 0;
	thrust::complex<double> YR = 0, YG = 0, YB = 0;

	int threadID = blockIdx.x * blockDim.x + threadIdx.x;

	if (threadID > loopSize * loopSize - 1)
		return;

	int size = ceil((1.0 * loopSize * loopSize) / numOfThreads);
	if (size == 0)
		size = 1;

	int local_start = threadID * size;
	int local_end = (threadID + 1) * size - 1;

	if (local_start > loopSize * loopSize)
		return;

	for (int i = local_start; i <= local_end; i++) {

		p = i / (loopSize);
		q = i % loopSize;
		S1 = S2 = S3 = S4 = S5 = S6 = 0;

		for (int i = 1; i <= m; i++) {
			p_index = p * m + i - 1;
			for (int j = 1; j <= 2 * i - 1; j++) {
				z1 = 4 * i - 1 - j;
				z2 = 4 * i - 2 + j;
				z3 = 8 * i - 3 - j;
				i_index = (2 * (i - 1)) * (2 * (i - 1));
				q_index = (2 * (i - 1)) * (2 * (i - 1)) * (max + 1)
						+ (q) * ((2 * (i - 1) + 1) * S);

				YR = fR[i_index + (j - 1)] * Iq[q_index + (j - 1)]
						+ fR[i_index + z1 - 1] * Iq[q_index + (z1 - 1)]
						+ fR[i_index + z2 - 1] * Iq[q_index + z2 - 1]
						+ fR[i_index + z3 - 1] * Iq[q_index + z3 - 1];
				YG = fG[i_index + (j - 1)] * Iq[q_index + (j - 1)]
						+ fG[i_index + z1 - 1] * Iq[q_index + (z1 - 1)]
						+ fG[i_index + z2 - 1] * Iq[q_index + z2 - 1]
						+ fG[i_index + z3 - 1] * Iq[q_index + z3 - 1];
				YB = fB[i_index + (j - 1)] * Iq[q_index + (j - 1)]
						+ fB[i_index + z1 - 1] * Iq[q_index + (z1 - 1)]
						+ fB[i_index + z2 - 1] * Iq[q_index + z2 - 1]
						+ fB[i_index + z3 - 1] * Iq[q_index + z3 - 1];

				S1 = S1 + (YR) * Ip[p_index];
				S2 = S2 + (YG) * Ip[p_index];
				S3 = S3 + (YB) * Ip[p_index];

				S4 = S4 + (YR) * conj(Ip[p_index]);
				S5 = S5 + (YG) * conj(Ip[p_index]);
				S6 = S6 + (YB) * conj(Ip[p_index]);
			}
		}

		SR[p * (max + 1) + q] = pii * S1;
		SG[p * (max + 1) + q] = pii * S2;
		SB[p * (max + 1) + q] = pii * S3;
		H1R[p * (max + 1) + q] = pii * S4;
		H1G[p * (max + 1) + q] = pii * S5;
		H1B[p * (max + 1) + q] = pii * S6;
	}
}

__global__ void QuaternionReconstructionKernel(double * r, double * Theta1,
		thrust::complex<double> *SR, thrust::complex<double> *SB,
		thrust::complex<double> *SG, thrust::complex<double> *H1R,
		thrust::complex<double> *H1B, thrust::complex<double> *H1G, double *h_1,
		double *h_2, double *h_3, int max, int image_size, int numOfThreads) {

	int threadID = blockIdx.x * blockDim.x + threadIdx.x;

	int threadLoad = ceil((1.0 * image_size * image_size) / numOfThreads);
	if (threadLoad == 0)
		threadLoad = 1;

	int local_start = threadID * threadLoad;
	int local_end = (threadID + 1) * threadLoad - 1;

	if (local_start > image_size * image_size - 1)
		return;

	if (local_end > image_size * image_size - 1)
		local_end = image_size * image_size - 1;

	double PI = 3.1415926535897;
	thrust::complex<double> sum3 = 0.0, sum2 = 0.0, sum1 = 0.0, s1 = 0.0, s2 =
			0.0, s3 = 0.0;

	int pmax, qmax;
	pmax = qmax = max;
	thrust::complex<double> ii(0, 1);
	thrust::complex<double> R1, R2;

	int p_index = 0;
	int temp_index;
	for (int it = local_start; it <= local_end; it++) {

		int i = it / (image_size) + 1;
		int j = it % (image_size) + 1;

		for (int p = 1; p <= pmax; p++) {
			temp_index = (i - 1) * image_size + (j - 1);
			R1 = exp(p * 2.0 * ii * PI * pow(r[temp_index], 2.0));
			R2 = exp(p * Theta1[temp_index] * ii);

			p_index = p * (max + 1);
			s1 = s1 + (SR[p_index]) * (R1) + conj(SR[p_index]) * conj(R1)
					+ (SR[p]) * (R2) + conj(SR[p] * (R2));
			s2 = s2 + (SG[p_index]) * (R1) + conj(SG[p_index]) * conj(R1)
					+ (SG[p]) * (R2) + conj(SG[p] * (R2));
			s3 = s3 + (SB[p_index]) * (R1) + conj(SB[p_index]) * conj(R1)
					+ (SB[p]) * (R2) + conj(SB[p] * (R2));

			for (int q = 1; q <= qmax; q++) {

				R2 = exp(q * Theta1[temp_index] * ii);
				sum1 = sum1 + SR[p_index + q] * R1 * R2
						+ conj(SR[p_index + q] * R1 * R2)
						+ H1R[p_index + q] * conj(R1) * (R2)
						+ conj(H1R[p_index + q]) * R1 * conj(R2);
				sum2 = sum2 + SG[p_index + q] * R1 * R2
						+ conj(SG[p_index + q] * R1 * R2)
						+ H1G[p_index + q] * conj(R1) * (R2)
						+ conj(H1G[p_index + q]) * R1 * conj(R2);
				sum3 = sum3 + SB[p_index + q] * R1 * R2
						+ conj(SB[p_index + q] * R1 * R2)
						+ H1B[p_index + q] * conj(R1) * (R2)
						+ conj(H1B[p_index + q]) * R1 * conj(R2);
			}
		}
		h_3[temp_index] = (sum3 + s3 + SB[0]).real();
		h_2[temp_index] = (sum2 + s2 + SG[0]).real();
		h_1[temp_index] = (sum1 + s1 + SR[0]).real();

		sum1 = sum2 = sum3 = 0.0;
		s1 = s2 = s3 = 0;
	}
}
vector<double> interker(vector<double> x) {
	vector<double> hx(SIZE);
	double Absx[SIZE];
	for (unsigned i = 0; i < x.size(); i++) {
		Absx[i] = abs(x[i]);
	}

	for (unsigned i = 0; i < x.size(); i++) {
		if (Absx[i] <= 1)
			hx[i] = (3.0 / 2.0) * pow(Absx[i], 3)
					- (5.0 / 2.0) * pow(Absx[i], 2) + 1;
		else if (Absx[i] <= 2)
			hx[i] = (-1.0 / 2) * pow(Absx[i], 3) + (5.0 / 2.0) * pow(Absx[i], 2)
					- 4 * Absx[i] + 2;
		else
			hx[i] = 0;
	}
	return hx;
}

bool computeIpIq() {
	struct timeval startAccesstime, endAccesstime;

	int Ki[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ki[i] = (2 * i + 1) * S;
	}

	double Ri[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ri[i] = (2 * i + 1) / (2.0 * M);
	}

	double deltaR = 1.0 / M;
	double U[M + 1];
	U[0] = Ri[0] - (deltaR / 2.0);

	for (int i = 1; i <= M; i++) {
		U[i] = Ri[i - 1] + deltaR / 2.0;
	}

	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		Theta1.push_back(x);
	}

	double PI = 3.1415926535897;

	for (int i = 1; i <= M; i++) {
		for (int n = 0; n < Ki[i - 1]; n++) {
			Theta1[i - 1][n] = (n + 0.5) * (2 * PI / Ki[i - 1]);
		}
	}

	//the upper and lower limit of the sectors angles
	double deltaTheta;

	vector<vector<double> > V;
	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		V.push_back(x);
	}

	for (int i = 0; i < M; i++) {
		deltaTheta = 2 * PI / Ki[i];
		V[i][0] = Theta1[i][0] - (deltaTheta / 2.0);
		for (int n = 1; n < Ki[i] + 1; n++) {
			V[i][n] = Theta1[i][n - 1] + (deltaTheta / 2.0);
		}
	}

	double r[M];
	for (int i = 0; i < M; i++)
		r[i] = Ri[i];

	gettimeofday(&startAccesstime, NULL);
	int pmax = MAX;
	int qmax = MAX;

	std::complex<double> I3(0, 1);
	std::complex<double> i1neg(0, -1);

#pragma omp parallel for private(I3) num_threads(numOfThreadsCPU)
	for (int p = 0; p <= pmax; p++) {
		for (int i = 1; i <= M; i++) {
			if (p == 0)
				I3 = 0.5 * (pow(U[i], 2.0) - pow(U[i - 1], 2.0));
			else
				I3 = (exp(p * 2.0 * i1neg * PI * pow(U[i], 2.0))
						- exp(p * 2.0 * i1neg * PI * pow(U[i - 1], 2.0)))
						/ (p * 4.0 * PI * i1neg);
			Ip[p][i - 1] = I3;
		}
	}

	std::complex<double> I(0, 0);
	std::complex<double> i1pos(0, 1);

	for (int i = 0; i < M; i++) {
		v2d x(MAX + 1, v1d(S * (2 * i + 1)));
		Iq.push_back(x);
	}

	//% %%%%%%%%%%%%%%%%%%%%%%I(q)%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#pragma omp parallel for private(I) num_threads(numOfThreadsCPU)
	for (int q = 0; q <= qmax; q++) {
		for (int i = 1; i <= M; i++) {
			for (int j = 1; j <= Ki[i - 1]; j++) {
				if (q == 0)
					I = V[i - 1][j] - V[i - 1][j - 1];
				else
					I = (i1pos / (double) q)
							* (exp(-1.0 * q * i1pos * V[i - 1][j])
									- exp(-1.0 * q * i1pos * V[i - 1][j - 1]));
				Iq[i - 1][q][j - 1] = I;
			}
		}
	}

	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "I_p & I_q time = " << totTime << " seconds." << endl;
	return true;
}

bool transferImageIndependentDataToDevice() {
	struct timeval startAccesstime, endAccesstime;
	gettimeofday(&startAccesstime, NULL);

	size_t a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * M);
	cudaMalloc((void **) &d_Ip, a_size);

	for (int i = 0; i < (MAX + 1); ++i) {
		cudaMemcpyAsync(d_Ip + i * (M), &Ip[i][0],
				sizeof(thrust::complex<double>) * M, cudaMemcpyHostToDevice);
	}

	a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * SIZE * SIZE);

	cudaMalloc((void **) &d_Iq, a_size);
	for (int k = 0, currentPosition = 0; k < M; k++) {
		int row_size = ((2 * (k) + 1) * S);
		for (int i = 0; i < (MAX + 1);
				++i, currentPosition += ((2 * (k) + 1) * S)) {
			cudaMemcpyAsync(d_Iq + currentPosition, &Iq[k][i][0],
					sizeof(thrust::complex<double>) * row_size,
					cudaMemcpyHostToDevice);
		}
	}
	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	cudaError err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << __LINE__ << " " << cudaGetErrorString(err) << endl;
		exit(-1);
	}
	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image independent data transfer HtoD time = " << totTime
			<< " seconds." << endl;

	return true;
}

void GPU_Quaternion_Moment_Computation() {

	struct timeval startAccesstime, endAccesstime;

	int Ki[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ki[i] = (2 * i + 1) * S;
	}

	double PI = 3.1415926535897;

	for (int i = 1; i <= M; i++) {
		for (int n = 0; n < Ki[i - 1]; n++) {
			Theta1[i - 1][n] = (n + 0.5) * (2 * PI / Ki[i - 1]);
		}
	}

	double Ri[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ri[i] = (2 * i + 1) / (2.0 * M);
	}
	double r[M];
	for (int i = 0; i < M; i++)
		r[i] = Ri[i];

	//compute xi,yi
	int s[SIZE];
	for (int i = 0; i < SIZE; i++)
		s[i] = i + 1;

	double xi[SIZE];
	for (int i = 0; i < SIZE; i++)
		xi[i] = (2.0 * s[i] - SIZE - 1) / SIZE;

	double yi[SIZE];
	for (int i = 0; i < SIZE; i++)
		yi[i] = (2.0 * s[i] - SIZE - 1) / SIZE;

	double del = 2.0 / SIZE;   //% the pixel width
	const int N = SIZE;

	gettimeofday(&startAccesstime, NULL);

	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		fR.push_back(x);
		fB.push_back(x);
		fG.push_back(x);
	}

#pragma omp parallel for schedule(dynamic,1) num_threads(numOfThreadsCPU)
	for (unsigned i = 1; i <= M; i++) {
		for (unsigned j = 1; j <= Ki[i - 1]; j++) {

			int k1 = ceil(r[i - 1] * cos(Theta1[i - 1][j - 1]) / del) + N / 2;
			int k2 = ceil(r[i - 1] * sin(Theta1[i - 1][j - 1]) / del) + N / 2;

			int lSize = (k1 + 2) - (k1 - 1) + 1;
			int mSize = (k2 + 2) - (k2 - 1) + 1;

			std::vector<int> l(lSize);
			std::vector<int> m(mSize);
			std::vector<int> inx(lSize);
			std::vector<int> iny(mSize);

			for (int i = 0, count = k1 - 1; i < lSize; i++) {
				l[i] = count++;
				inx[i] = l[i];
			}

			for (int i = 0, count = k2 - 1; i < mSize; i++) {
				m[i] = count++;
				iny[i] = m[i];
			}

			// construct image matrix
			for (int i = 0; i < lSize; i++) {
				if (l[i] > N) {
					inx[i] = l[i] % N;
					l[i] = N - (l[i] % N) + 1;
					continue;
				}
				if (l[i] < 1) {
					inx[i] = N - abs(l[i]);
					l[i] = N - ((l[i] + N) - 1);
				}

			}

			for (int i = 0; i < mSize; i++) {
				if (m[i] > N) {
					iny[i] = m[i] % N;
					m[i] = N - (m[i] % N) + 1;
					continue;
				}
				if (m[i] < 1) {
					iny[i] = N - abs(m[i]);
					m[i] = N - ((m[i] + N) - 1);
				}

			}
			//%construct the image 4*4 matrix
			vector<vector<double> > Imatrix1(lSize, vector<double>(mSize));
			vector<vector<double> > Imatrix2(lSize, vector<double>(mSize));
			vector<vector<double> > Imatrix3(lSize, vector<double>(mSize));

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix1[i][j] = red[l[i] - 1][m[j] - 1];
				}
			}

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix2[i][j] = green[l[i] - 1][m[j] - 1];
				}
			}

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix3[i][j] = blue[l[i] - 1][m[j] - 1];
				}
			}
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//% construct kernel
			std::vector<double> xl(lSize);

			for (int i = 0; i < lSize; i++)
				xl[i] = xi[inx[i] - 1];

			std::vector<double> ym(mSize);

			for (int i = 0; i < mSize; i++)
				ym[i] = yi[iny[i] - 1];

			// % compute the kernel function using 'interker' function
			std::vector<double> tempXl(lSize);

			for (int ww = 0; ww < lSize; ww++) {
				tempXl[ww] = (r[i - 1] * cos(Theta1[i - 1][j - 1]) - xl[ww])
						/ del;
			}

			std::vector<double> h1(lSize);

			vector<double> hx = interker(tempXl);  //% for x
			for (int i = 0; i < lSize; i++)
				h1[i] = hx[i];

			std::vector<double> tempYm(mSize);

			for (int ww = 0; ww < mSize; ww++)
				tempYm[ww] = (r[i - 1] * sin(Theta1[i - 1][j - 1]) - ym[ww])
						/ del;

			std::vector<double> h2(mSize);

			hx = interker(tempYm);  //% for x
			for (int i = 0; i < lSize; i++)
				h2[i] = hx[i];

			vector<vector<double> > hh(lSize, vector<double>(mSize));
			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++)
					hh[i][j] = h1[i] * h2[j];

			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//% 2D convolution between the kernel and the image matrix
			vector<vector<double> > P1(lSize, vector<double>(mSize));
			vector<vector<double> > P2(lSize, vector<double>(mSize));
			vector<vector<double> > P3(lSize, vector<double>(mSize));

			double sum1 = 0, sum2 = 0, sum3 = 0;
			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P1[i][j] = Imatrix1[i][j] * hh[i][j];
					sum1 += P1[i][j];
				}

			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P2[i][j] = Imatrix2[i][j] * hh[i][j];
					sum2 += P2[i][j];
				}

			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P3[i][j] = Imatrix3[i][j] * hh[i][j];
					sum3 += P3[i][j];
				}
			fR[i - 1][j - 1] = sum1;
			fG[i - 1][j - 1] = sum2;
			fB[i - 1][j - 1] = sum3;

			//clearing area
			P1.clear();
			P1.clear();
			P3.clear();
			hh.clear();
			h2.clear();
			tempYm.clear();
			h1.clear();
			tempXl.clear();
			ym.clear();
			xl.clear();
			Imatrix3.clear();
			Imatrix2.clear();
			Imatrix1.clear();
			iny.clear();
			inx.clear();
			m.clear();
			l.clear();

		}
	}
	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Interpolation time = " << totTime << " seconds." << endl;

	//%%%%%%%%%%%%%%%%%%%%%%%%%QPCET _  Moment Computed.................
	cout << "\n----------------------------------------" << endl;
	cout << "Quat moment computation started on GPU...." << endl;
	gettimeofday(&startAccesstime, NULL);

	//fR
	double * d_fR;
	size_t a_size = sizeof(double) * size_t(SIZE * SIZE);
	cudaMalloc((void **) &d_fR, a_size);

	double * d_fB;
	cudaMalloc((void **) &d_fB, a_size);

	double * d_fG;
	cudaMalloc((void **) &d_fG, a_size);

	for (int i = 0, currentPosition = 0; i < M; i++) {
		cudaMemcpy(d_fR + currentPosition, &fR[i][0],
				sizeof(double) * ((2 * (i) + 1) * S), cudaMemcpyHostToDevice);
		cudaMemcpy(d_fG + currentPosition, &fG[i][0],
				sizeof(double) * ((2 * (i) + 1) * S), cudaMemcpyHostToDevice);
		cudaMemcpy(d_fB + currentPosition, &fB[i][0],
				sizeof(double) * ((2 * (i) + 1) * S), cudaMemcpyHostToDevice);
		currentPosition += ((2 * (i) + 1) * S);
	}

	thrust::complex<double> * d_SG;
	thrust::complex<double> * d_SB;
	thrust::complex<double> * d_SR;

	thrust::complex<double> * d_H1G;
	thrust::complex<double> * d_H1B;
	thrust::complex<double> * d_H1R;

	a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * (MAX + 1));
	cudaMalloc((void **) &d_SG, a_size);
	cudaMemset(d_SG, 0, a_size);

	cudaMalloc((void **) &d_SB, a_size);
	cudaMemset(d_SB, 0, a_size);

	cudaMalloc((void **) &d_SR, a_size);
	cudaMemset(d_SR, 0, a_size);

	cudaMalloc((void **) &d_H1G, a_size);
	cudaMemset(d_H1G, 0, a_size);

	cudaMalloc((void **) &d_H1B, a_size);
	cudaMemset(d_H1B, 0, a_size);

	cudaMalloc((void **) &d_H1R, a_size);
	cudaMemset(d_H1R, 0, a_size);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	cudaError err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << __LINE__ << " " << cudaGetErrorString(err) << endl;
		exit(-1);
	}

	gettimeofday(&endAccesstime, NULL);

	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Quater. moment GPU data transfer HtoD time = " << totTime << endl;

	gettimeofday(&startAccesstime, NULL);

	//GPU only
	QuaternionMomentKernel<<<noOfBlocks, noOfThreads>>>(d_Ip, d_Iq, d_fR,
			d_fB, d_fG, d_SR, d_SB, d_SG, d_H1R, d_H1B, d_H1G, M, MAX,
			totNoOfThreads);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << cudaGetErrorString(err) << endl;
		exit(-1);
	}

	gettimeofday(&endAccesstime, NULL);

	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Quater moment computation on a single GPU time = " << totTime
			<< " seconds. # of GPU threads: " << totNoOfThreads << endl;

	gettimeofday(&startAccesstime, NULL);

	h_SB = (thrust::complex<double>*) malloc(
			((MAX + 1) * (MAX + 1)) * sizeof(thrust::complex<double>));
	a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * (MAX + 1));
	cudaMemcpy(h_SB, d_SB, a_size, cudaMemcpyDeviceToHost);

	h_SG = (thrust::complex<double>*) malloc(
			((MAX + 1) * (MAX + 1)) * sizeof(thrust::complex<double>));
	cudaMemcpy(h_SG, d_SG, a_size, cudaMemcpyDeviceToHost);

	h_SR = (thrust::complex<double>*) malloc(
			((MAX + 1) * (MAX + 1)) * sizeof(thrust::complex<double>));
	cudaMemcpy(h_SR, d_SR, a_size, cudaMemcpyDeviceToHost);

	h_H1B = (thrust::complex<double>*) malloc(
			((MAX + 1) * (MAX + 1)) * sizeof(thrust::complex<double>));
	a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * (MAX + 1));
	cudaMemcpy(h_H1B, d_H1B, a_size, cudaMemcpyDeviceToHost);

	h_H1G = (thrust::complex<double>*) malloc(
			((MAX + 1) * (MAX + 1)) * sizeof(thrust::complex<double>));
	cudaMemcpy(h_H1G, d_H1G, a_size, cudaMemcpyDeviceToHost);

	h_H1R = (thrust::complex<double>*) malloc(
			((MAX + 1) * (MAX + 1)) * sizeof(thrust::complex<double>));
	cudaMemcpy(h_H1R, d_H1R, a_size, cudaMemcpyDeviceToHost);

	cudaDeviceSynchronize();

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Quater. moment GPU data transfer DtoH time = " << totTime
			<< " seconds." << endl;
	cout << "Quaternion moment is copied from device to host." << endl;

	cudaFree(d_Ip);
	cudaFree(d_Iq);
	cudaFree(d_fB);
	cudaFree(d_fG);
	cudaFree(d_fR);
	cudaFree(d_SB);
	cudaFree(d_SG);
	cudaFree(d_SR);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << "GPU memory de-allocation: " << cudaGetErrorString(err) << endl;
		exit(-1);
	}
}

void errorComputing() {

	double x[SIZE];
	for (int i = 1; i <= SIZE; i++) {
		x[i - 1] = (2 * i - SIZE - 1) / (double) SIZE;
	}
	vector<vector<double> > r(SIZE, vector<double>(SIZE));

	for (int i = 0; i < SIZE; i++)
		for (int j = 0; j < SIZE; j++) {
			r[i][j] = sqrt(pow(x[i], 2) + pow(x[j], 2));
		}
	//%%%%%%%%%%NIRE Error%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	double sun1 = 0.0, sun11 = 0.0, sun22 = 0.0, sun222 = 0.0, sun2 = 0.0,
			sun3 = 0.0;
	int temp_index = 0;
	for (int i = 1; i <= SIZE; i++) {
		for (int j = 1; j <= SIZE; j++) {
			temp_index = (i - 1) * SIZE + (j - 1);
			if (r[i - 1][j - 1] <= 1) {
				sun2 = sun2 + pow(blue[i - 1][j - 1], 2.0);
				sun3 = sun3 + pow(h_3[temp_index] - blue[i - 1][j - 1], 2.0);
				// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				sun1 = sun1 + pow(red[i - 1][j - 1], 2.0);
				sun11 = sun11 + pow(h_1[temp_index] - red[i - 1][j - 1], 2.0);
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				sun22 = sun22 + pow(green[i - 1][j - 1], 2.0);
				sun222 = sun222
						+ pow(h_2[temp_index] - green[i - 1][j - 1], 2.0);
			} else {
				h_3[temp_index] = h_2[temp_index] = h_1[temp_index] = 255;
			}

		}
	}
	cout << endl;
	double dif3 = (sun3 / sun2);
	double dif2 = (sun222 / sun22);
	double dif1 = (sun11 / sun1);
	double dif = (dif2 + dif3 + dif1) / 3.0;
	cout << "Difference between original and reconstructed image = " << dif
			<< endl;
}

bool saveReconstructedImage(bool showImage = false) {
	int temp_index = 0;
	int size = SIZE;

	cv::Mat mat_red(size, size, CV_64FC1);
	cv::Mat mat_green(size, size, CV_64FC1);
	cv::Mat mat_blue(size, size, CV_64FC1);

	for (int i = 0; i < mat_red.rows; ++i)
		for (int j = 0; j < mat_red.cols; ++j) {
			temp_index = i * SIZE + j;
			mat_red.at<double>(i, j) = h_1[temp_index];
			mat_green.at<double>(i, j) = h_2[temp_index];
			mat_blue.at<double>(i, j) = h_3[temp_index];
		}

	Mat temp;
	vector<Mat> channels;
	channels.push_back(mat_blue);
	channels.push_back(mat_green);
	channels.push_back(mat_red);

	merge(channels, temp);
	Mat final_reconstructed_image;
	temp.convertTo(final_reconstructed_image, CV_8UC1);
	String imName = "reconstructed-" + std::to_string(MAX) + "-"
			+ imageFileName;
	imwrite(imName, final_reconstructed_image);
	if (showImage) {
		imshow(">>Reconst. Image", final_reconstructed_image);
		waitKey(0);
	}
	cout << "The reconstructed image is saved on disk with file name: "
			<< imName << endl;
	return true;
}

bool readImage(const String imageName) {

	ifstream image_file(imageName.c_str());

	if (!image_file.good()) {
		cerr << "Error: the image with name \"" << imageName
				<< "\" doesn't exist." << endl;
		return false;
	}
	cout << "Reading image: " << imageName << "......";
	Mat image = cv::imread(imageName);
	if (image.channels() != 3) {
		cerr << "Error: only images with three channels are supported" << endl;
		return false;
	}
	int tempRow = -1;
	int tempCol = -1;
	if (image.rows != SIZE || image.cols != SIZE) {
		tempRow = image.rows;
		tempCol = image.cols;
		resize(image, image, Size(SIZE, SIZE));
	}

	Mat bgr[3];   //destination array
	split(image, bgr);   //split source

	//Note: OpenCV uses BGR color order
	imwrite("blue.png", bgr[0]);   //blue channel
	imwrite("green.png", bgr[1]);   //green channel
	imwrite("red.png", bgr[2]);   //red channel

	vector<vector<int> > temp_blue(bgr[0].rows, vector<int>(bgr[0].cols));
	vector<vector<int> > temp_green(bgr[1].rows, vector<int>(bgr[1].cols));
	vector<vector<int> > temp_red(bgr[2].rows, vector<int>(bgr[2].cols));

	for (int i = 0; i < bgr[0].rows; i++)
		for (int j = 0; j < bgr[0].cols; j++) {
			if ((int) bgr[0].at<char>(i, j) >= 0) {
				temp_blue[i][j] = (int) bgr[0].at<char>(i, j);
			} else {
				temp_blue[i][j] = 256 + (int) bgr[0].at<char>(i, j);
			}
			if ((int) bgr[1].at<char>(i, j) >= 0) {
				temp_green[i][j] = (int) bgr[1].at<char>(i, j);
			} else {
				temp_green[i][j] = 256 + (int) bgr[1].at<char>(i, j);
			}

			if ((int) bgr[2].at<char>(i, j) >= 0) {
				temp_red[i][j] = (int) bgr[2].at<char>(i, j);
			} else {
				temp_red[i][j] = 256 + (int) bgr[2].at<char>(i, j);
			}
		}

	blue = temp_blue;
	green = temp_green;
	red = temp_red;

	cout << "Done!" << endl;

	if (tempCol != -1)
		cout << "The input image size is:" << tempRow << "x" << tempCol
				<< ". The image is resized to this dimensions: " << SIZE << "x"
				<< SIZE << "." << endl;

	return true;
}

static void show_usage(std::string name) {
	std::cerr << "Usage: " << name << " <option(s)>: \n"
			<< "\t--image\t set the input file image\n"
			<< "\t--threads\t set the used number of CPU threads.\n"
			<<"\t--gpu-threads\t set the number of threads per GPU block.\n"
			<<"\t--gpu-blocks\t set the number of GPU blocks."
			<< std::endl;
}

bool set_inputs(int argc, const char * argv[]) {
	if (argc < 3) {
		show_usage(argv[0]);
		return false;
	}

	for (int i = 1; i < argc; ++i) {
		std::string arg = argv[i];
		if (arg == "--image") {
			imageFileName = argv[i + 1];
			i++;
			if (!readImage(imageFileName)) {
				return false;
			}
		} else if (arg == "--threads") {
			numOfThreadsCPU = atoi(argv[i + 1]);
			i++;
			if (numOfThreadsCPU < 2) {
				std::cerr << "The number of CPU threads should be more than 1."
						<< endl;
				return false;
			}
			if (numOfThreadsCPU > (MAX + 1) * (MAX + 1)) {
				numOfThreadsCPU = (MAX + 1) * (MAX + 1);
				std::cerr << "The number of CPU threads is adjusted to "
						<< numOfThreadsCPU
						<< ". The maximum allowed number of CPU thread is (moment order +1)^2 = ("
						<< MAX << " + 1)^2." << endl;
			}
			cout << "The number of used CPU threads is set to: "
					<< numOfThreadsCPU << " threads." << endl;
		} else if (arg == "--gpu-threads") {
			noOfThreads = atoi(argv[i + 1]);
			i++;
			if ((noOfThreads % 32) != 0 || noOfThreads < 32) {
				std::cerr
						<< "The number of GPU threads per block should be a multiple of 32."
						<< endl;
				return false;
			}
			if (noOfThreads > 1024) {
				noOfThreads = 1024;
				std::cerr
						<< "The number of GPU threads per block is adjusted to "
						<< noOfThreads
						<< ". The maximum allowed number of GPU threads per block is 1024 thread."
						<< endl;
			}
			cout << "The number of GPU threads per block is set to: "
					<< noOfThreads << " blocks." << endl;
		} else if (arg == "--gpu-blocks") {
			noOfBlocks = atoi(argv[i + 1]);

			cudaDeviceProp prop;
			cudaGetDeviceProperties(&prop, 0);
			int numOfMultiProcessors = prop.multiProcessorCount;
			i++;
			if ((noOfBlocks % numOfMultiProcessors) != 0
					|| noOfBlocks < numOfMultiProcessors) {
				std::cerr << "The number of GPU blocks should be a multiple of "
						<< numOfMultiProcessors
						<< ", as the number of SMs is:  "
						<< numOfMultiProcessors << endl;
				return false;
			}
			cout << "The number of GPU blocks is set to: " << noOfBlocks
					<< " threads. The number of GPU SMs is: "
					<< numOfMultiProcessors << endl;
		} else {
			std::cerr << argv[i] << " is an invalid argument." << endl;
			return false;
		}
	}
	return true;
}

void GPU_Image_Reconstruct_Computation() {

	double x[SIZE];

	for (int i = 1; i <= SIZE; i++) {
		x[i - 1] = ((2 * i) - SIZE - 1) / (double) SIZE;
	}

	vector<vector<double> > Theta1(SIZE, vector<double>(SIZE));
	vector<vector<double> > r(SIZE, vector<double>(SIZE));

	for (int i = 0; i < SIZE; i++)
		for (int j = 0; j < SIZE; j++) {
			r[i][j] = sqrt(pow(x[i], 2) + pow(x[j], 2));
			Theta1[i][j] = std::atan2(x[j], x[i]);
		}

	struct timeval startAccesstime, endAccesstime;

	int numOfThreads = noOfBlocks * noOfThreads;
	cout << "\n----------------------------------------" << endl;
	cout << "Quat image reconstruction computation started on GPU...." << endl;

	gettimeofday(&startAccesstime, NULL);

	//r
	double * d_r;
	size_t a_size = sizeof(double) * size_t(SIZE * SIZE);
	cudaMalloc((void **) &d_r, a_size);
	//theta1
	double * d_Theta1;
	a_size = sizeof(double) * size_t(SIZE * SIZE);
	cudaMalloc((void **) &d_Theta1, a_size);

	for (int i = 0, currentPosition = 0; i < SIZE; i++) {
		cudaMemcpyAsync(d_r + currentPosition, &r[i][0], sizeof(double) * SIZE,
				cudaMemcpyHostToDevice);
		cudaMemcpyAsync(d_Theta1 + currentPosition, &Theta1[i][0],
				sizeof(double) * SIZE, cudaMemcpyHostToDevice);
		currentPosition += SIZE;
	}

	a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * (MAX + 1));

	thrust::complex<double> * d_SG;
	cudaMalloc((void **) &d_SG, a_size);
	cudaMemcpy(d_SG, h_SG, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_SB;
	cudaMalloc((void **) &d_SB, a_size);
	cudaMemcpy(d_SB, h_SB, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_SR;
	cudaMalloc((void **) &d_SR, a_size);
	cudaMemcpy(d_SR, h_SR, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_H1G;
	cudaMalloc((void **) &d_H1G, a_size);
	cudaMemcpy(d_H1G, h_H1G, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_H1B;
	cudaMalloc((void **) &d_H1B, a_size);
	cudaMemcpy(d_H1B, h_H1B, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_H1R;
	cudaMalloc((void **) &d_H1R, a_size);
	cudaMemcpy(d_H1R, h_H1R, a_size, cudaMemcpyHostToDevice);

	a_size = sizeof(double) * size_t(SIZE * SIZE);
	double * d_h1;
	cudaMalloc((void **) &d_h1, a_size);

	double * d_h2;
	cudaMalloc((void **) &d_h2, a_size);

	double * d_h3;
	cudaMalloc((void **) &d_h3, a_size);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	cudaError err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << "GPU memory allocation: " << cudaGetErrorString(err) << endl;
		exit(-1);
	}

	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Reconst on GPU: data transfer HtoD time  = " << totTime
			<< endl;

	gettimeofday(&startAccesstime, NULL);
	QuaternionReconstructionKernel<<<noOfBlocks, noOfThreads>>>(d_r, d_Theta1,
			d_SR, d_SB, d_SG, d_H1R, d_H1B, d_H1G, d_h1, d_h2, d_h3,
			MAX, SIZE, numOfThreads);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << "Reconst kernel: " << cudaGetErrorString(err) << endl;
		exit(-1);
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image reconst on a single GPU running time  = " << totTime
			<< " seconds." << " # of GPU threads: " << totNoOfThreads << endl;

	gettimeofday(&startAccesstime, NULL);

	a_size = sizeof(double) * size_t(SIZE * SIZE);
	h_1 = (double*) malloc(a_size);
	h_2 = (double*) malloc(a_size);
	h_3 = (double*) malloc(a_size);

	cudaMemcpy(h_1, d_h1, a_size, cudaMemcpyDeviceToHost);
	cudaMemcpy(h_2, d_h2, a_size, cudaMemcpyDeviceToHost);
	cudaMemcpy(h_3, d_h3, a_size, cudaMemcpyDeviceToHost);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	err = cudaGetLastError();

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Reconst on GPU: data transfer DtoH time  = " << totTime
			<< " seconds." << endl;

	cudaFree(d_SB);
	cudaFree(d_SG);
	cudaFree(d_SR);
	cudaFree(d_H1B);
	cudaFree(d_H1G);
	cudaFree(d_H1R);

//		ofstream myfile;
//		myfile.open("gpu_h1.txt");
//
//		for (int i = 0; i < SIZE; i++){
//			for (int j = 0; j < SIZE; j++) {
//				myfile << h_h1[i * SIZE + j]<<"," ;
//				h_1 [i * SIZE + j] = h_h1[i * SIZE + j];
//				h_2 [i * SIZE + j] = h_h2[i * SIZE + j];
//				h_3 [i * SIZE + j] = h_h3[i * SIZE + j];
//
//			}
//			myfile<<endl;
//		}
//		myfile.close();
}

int main(int argc, const char **argv) {
	if (!set_inputs(argc, argv))
		return 0;

	totNoOfThreads = noOfBlocks * noOfThreads;
	omp_set_dynamic(0);
	omp_set_num_threads(numOfThreadsCPU);

#pragma omp parallel
	{
		int ID = omp_get_thread_num();
	}

	//image independent tasks
	computeIpIq();
	transferImageIndependentDataToDevice();

	//image dependent tasks
	GPU_Quaternion_Moment_Computation();
	GPU_Image_Reconstruct_Computation();

	//evaluating the moment accuracy numerically and visually
	errorComputing();
	saveReconstructedImage();
	cout << "All tasks are done!" << endl;

	return 0;
}
