/**
 * Copyright 1993-2012 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 */
#include <stdio.h>
#include<cmath>
#include <unistd.h>
#include <numeric>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <complex>
#include <omp.h>
#include<sstream>
#include<string>

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/opencv.hpp>
#include <opencv2/imgproc.hpp>

#include <sys/time.h>
#include <vector>
#include <fstream>
#include <thrust/complex.h>

using namespace std;
using namespace cv;

#include <cuda.h>

int numOfThreadsCPU = 16;

int noOfBlocks = 10;
int noOfThreads = 512;
int totNoOfThreads = 0;
char * fileName;

const int SIZE = 256;
#define MAX 20

const int M = SIZE / 2;
const int S = 4;

vector<vector<int> > blue;
vector<vector<int> > green;
vector<vector<int> > red;
String imageFileName;

int numOfDevices = 1;
int numOfGPUs = 1;
bool usingCPU = false;

//total should be M/2
//256 9,9,9,9,27
//512 {30,30,30,30,8}
//1024 { 216,40 } //{60,60,60,60,16}

//int devicePowers[] = { 16, 16, 16, 16, 0 }; // { 14, 13, 13, 13, 10 }; //{ 14, 14, 14, 14, 7 };//{ 3, 3, 57 }; //{ 14, 14, 14, 14, 7 }; //size = 256 total 63

vector<int> devicePowers;
//Iq declaration
typedef vector<std::complex<double> > v1d;
typedef vector<v1d> v2d;
typedef vector<v2d> v3d;
v3d Iq;

//Ip declaration
vector<vector<std::complex<double> > > Ip(M,
		vector<std::complex<double> >(MAX + 1));
vector<vector<double> > Theta1;
vector<vector<double> > fR, fG, fB;

//global moment data
thrust::complex<double>* h_SB;
thrust::complex<double>* h_SG;
thrust::complex<double>* h_SR;
thrust::complex<double>* h_H1B;
thrust::complex<double>* h_H1G;
thrust::complex<double>* h_H1R;

//image independent device data
typedef struct {
	unsigned start1, end1, start2, end2;
} DeviceWorkLoadInfo;
vector<DeviceWorkLoadInfo> works;
vector<thrust::complex<double> *> d_Ip(numOfGPUs);
vector<thrust::complex<double> *> d_Iq(numOfGPUs);
vector<DeviceWorkLoadInfo *> d_work(numOfGPUs);
vector<int> numOfSectorsPerGPU(numOfGPUs);

vector<unsigned *> d_sectorPerTrack(numOfGPUs);
vector<unsigned *> d_accumSectors(numOfGPUs);

double * h_1;
double * h_2;
double * h_3;

__device__ int numOfReplicas = 2;
int numOfReplicasCPU = 2;

//[i][0] means image size,
//[i][1] means moment order
//[i][2] first gpu share
//[i][3] second gpu share
//[i][4] third gpu share
//[i][5] fourth gpu share
//[i][6] cpu share
//[i][2] +[i][3] +[i][4]+[i][5]+[i][6] = SIZE/4

typedef struct {
	unsigned i, j;
} ThreadInfo;
unsigned partitionSizeRecommendationForMoment[9][7] = { { 256, 20, 15, 15, 15,
		15, 4 }, { 256, 40, 15, 15, 15, 15, 4 }, { 256, 60, 15, 15, 15, 15, 4 },
		{ 512, 20, 30, 30, 30, 30, 8 }, { 512, 40, 30, 30, 30, 30, 8 }, { 512,
				60, 30, 30, 30, 30, 8 }, { 1024, 20, 61, 61, 61, 61, 12 }, {
				1024, 40, 60, 60, 60, 60, 16 }, { 1024, 60, 60, 60, 60, 16 } };

//[i][0] means image size,
//[i][1] means moment order
//[i][2] first gpu share
//[i][3] second gpu share
//[i][4] third gpu share
//[i][5] fourth gpu share
//[i][6] cpu share
unsigned partitionSizeRecommendationForRecon[9][7] = { { 256, 20, 80, 80, 80,
		80, 1 }, { 256, 40, 75, 75, 75, 75, 1 }, { 256, 60, 80, 80, 80, 80, 1 },
		{ 512, 20, 100, 100, 100, 100, 1 }, { 512, 40, 100, 100, 100, 100, 1 },
		{ 512, 60, 100, 100, 100, 100, 1 }, { 1024, 20, 90, 90, 90, 90, 1 }, {
				1024, 40, 100, 100, 100, 100, 1 }, { 1024, 60, 100, 100, 100,
				100, 1 } };

inline void gpuAssert(cudaError_t code, const char *file, int line, bool abort =
		true) {
	if (code != cudaSuccess) {
		fprintf(stderr, "GPUassert: %s %s %d\n", cudaGetErrorString(code), file,
				line);
		if (abort)
			exit(code);
	}
}
#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }

__global__ void QuaternionMomentKernelMulti(thrust::complex<double> *Ip,
		thrust::complex<double> *Iq, double *fR, double *fB, double *fG,
		thrust::complex<double> *SR, thrust::complex<double> *SB,
		thrust::complex<double> *SG, thrust::complex<double> *H1R,
		thrust::complex<double> *H1B, thrust::complex<double> *H1G, int m,
		int max, int numOfThreads, DeviceWorkLoadInfo * workLoad,
		unsigned * sectorPerTrack, unsigned * accumSectors, int noOfReplicas) {

	int loopSize = max + 1;
	int totalMomentSize = loopSize * loopSize;

	int p, q;
	thrust::complex<double> S1 = 0, S2 = 0, S3 = 0, S4 = 0, S5 = 0, S6 = 0;
	thrust::complex<double> YR = 0, YG = 0, YB = 0;
	int z1, z2, z3;

	int threadID = blockIdx.x * blockDim.x + threadIdx.x;
	if (threadID > totalMomentSize * noOfReplicas - 1)
		return;

	int size = ceil((1.0 * totalMomentSize * noOfReplicas) / numOfThreads);
	if (size == 0)
		size = 1;

	int currentPartitionID;	// = threadID / totalMomentSize;
	int numOfRingsPerPartition = ((workLoad->end1 - workLoad->start1 + 1))
			/ noOfReplicas;
	if (numOfRingsPerPartition < 1)
		numOfRingsPerPartition = 1;
	int local_start = threadID * size;
	int local_end = local_start + size - 1;

	if (local_start > totalMomentSize * noOfReplicas - 1) {
		return;
	}
	if (local_end > totalMomentSize * noOfReplicas - 1) {
		local_end = totalMomentSize * noOfReplicas - 1;
	}
	int ringStart1, ringEnd1, ringEnd2, ringStart2;
	int temp = 0;

	int p_index = 0, q_index = 0;
	double pii = 7.0 / 22.0;

	for (int i = local_start; i <= local_end; i++) {

		currentPartitionID = i / totalMomentSize;

		ringStart1 = workLoad->start1
				+ currentPartitionID * numOfRingsPerPartition;
		ringEnd1 = workLoad->start1
				+ (currentPartitionID + 1) * numOfRingsPerPartition - 1;
		ringEnd2 = workLoad->end2 - currentPartitionID * numOfRingsPerPartition;
		ringStart2 = workLoad->end2
				- (currentPartitionID + 1) * numOfRingsPerPartition + 1;

		temp = i % totalMomentSize;
		p = temp / (loopSize);
		q = temp % loopSize;

		S1 = S2 = S3 = S4 = S5 = S6 = 0;

		int count = currentPartitionID * numOfRingsPerPartition;

		for (int i = ringStart1; i <= ringEnd1; i++, count++) {
			p_index = count * (max + 1) + p;
			for (int j = 1; j <= 2 * i - 1; j++) {
				z1 = 4 * i - 1 - j;
				z2 = 4 * i - 2 + j;
				z3 = 8 * i - 3 - j;

				q_index = accumSectors[count] * (max + 1)
						+ q * (sectorPerTrack[count]);
				YR = fR[accumSectors[count] + (j - 1)] * Iq[q_index + j - 1]
						+ fR[accumSectors[count] + z1 - 1]
								* Iq[q_index + z1 - 1]
						+ fR[accumSectors[count] + z2 - 1]
								* Iq[q_index + z2 - 1]
						+ fR[accumSectors[count] + z3 - 1]
								* Iq[q_index + z3 - 1];

				YG = fG[accumSectors[count] + j - 1] * Iq[q_index + j - 1]
						+ fG[accumSectors[count] + z1 - 1]
								* Iq[q_index + z1 - 1]
						+ fG[accumSectors[count] + z2 - 1]
								* Iq[q_index + z2 - 1]
						+ fG[accumSectors[count] + z3 - 1]
								* Iq[q_index + z3 - 1];

				YB = fB[accumSectors[count] + j - 1] * Iq[q_index + j - 1]
						+ fB[accumSectors[count] + z1 - 1]
								* Iq[q_index + z1 - 1]
						+ fB[accumSectors[count] + z2 - 1]
								* Iq[q_index + z2 - 1]
						+ fB[accumSectors[count] + z3 - 1]
								* Iq[q_index + z3 - 1];

				S1 = S1 + (YR) * Ip[p_index];
				S2 = S2 + (YG) * Ip[p_index];
				S3 = S3 + (YB) * Ip[p_index];

				S4 = S4 + (YR) * conj(Ip[p_index]);
				S5 = S5 + (YG) * conj(Ip[p_index]);
				S6 = S6 + (YB) * conj(Ip[p_index]);
			}
		}

		count = (workLoad->end1 - workLoad->start1 + 1) * 2 - (count);
		for (int i = ringStart2; i <= ringEnd2; i++, count++) {
			p_index = count * (max + 1) + p;
			for (int j = 1; j <= 2 * i - 1; j++) {
				z1 = 4 * i - 1 - j;
				z2 = 4 * i - 2 + j;
				z3 = 8 * i - 3 - j;

				q_index = accumSectors[count] * (max + 1)
						+ q * (sectorPerTrack[count]);

				YR = fR[accumSectors[count] + (j - 1)] * Iq[q_index + j - 1]
						+ fR[accumSectors[count] + z1 - 1]
								* Iq[q_index + z1 - 1]
						+ fR[accumSectors[count] + z2 - 1]
								* Iq[q_index + z2 - 1]
						+ fR[accumSectors[count] + z3 - 1]
								* Iq[q_index + z3 - 1];

				YG = fG[accumSectors[count] + j - 1] * Iq[q_index + j - 1]
						+ fG[accumSectors[count] + z1 - 1]
								* Iq[q_index + z1 - 1]
						+ fG[accumSectors[count] + z2 - 1]
								* Iq[q_index + z2 - 1]
						+ fG[accumSectors[count] + z3 - 1]
								* Iq[q_index + z3 - 1];

				YB = fB[accumSectors[count] + j - 1] * Iq[q_index + j - 1]
						+ fB[accumSectors[count] + z1 - 1]
								* Iq[q_index + z1 - 1]
						+ fB[accumSectors[count] + z2 - 1]
								* Iq[q_index + z2 - 1]
						+ fB[accumSectors[count] + z3 - 1]
								* Iq[q_index + z3 - 1];

				S1 = S1 + (YR) * Ip[p_index];
				S2 = S2 + (YG) * Ip[p_index];
				S3 = S3 + (YB) * Ip[p_index];

				S4 = S4 + (YR) * conj(Ip[p_index]);
				S5 = S5 + (YG) * conj(Ip[p_index]);
				S6 = S6 + (YB) * conj(Ip[p_index]);
			}
		}

		p_index = currentPartitionID * totalMomentSize + p * loopSize + q;
		SR[p_index] = pii * S1;
		SG[p_index] = pii * S2;
		SB[p_index] = pii * S3;

		H1R[p_index] = pii * S4;
		H1G[p_index] = pii * S5;
		H1B[p_index] = pii * S6;
	}
}

__global__ void Multi_GPU_CPU_QuaternionReconstructionKernel(double * r,
		double * Theta1, thrust::complex<double> *SR,
		thrust::complex<double> *SB, thrust::complex<double> *SG,
		thrust::complex<double> *H1R, thrust::complex<double> *H1B,
		thrust::complex<double> *H1G, double *h_1, double *h_2, double *h_3,
		int startIteration, int endIteration, int max, int image_size,
		int NumofThreads) {

	int totNumofIterations = endIteration - startIteration + 1;
	int size = ceil(1.0 * totNumofIterations / NumofThreads);

	if (size == 0)
		size = 1;
	int threadID = blockIdx.x * blockDim.x + threadIdx.x;

	unsigned localStart = startIteration + size * threadID;
	unsigned localEnd = localStart + size - 1;

	if (localStart > endIteration)
		return;
	if (localEnd > endIteration)
		localEnd = endIteration;

	double PI = 3.1415926535897;
	thrust::complex<double> i1(0, 1);
	thrust::complex<double> sum3 = 0.0, sum2 = 0.0, sum1 = 0.0, s1 = 0, s2 = 0,
			s3 = 0;

	int pmax, qmax;
	pmax = qmax = max;
	thrust::complex<double> ii(0, 1);
	thrust::complex<double> R1, R2;

	int p_index = 0;

	int temp_index;

	for (int it = localStart; it <= localEnd; it++) {
		int i = it / (image_size) + 1;
		int j = it % (image_size) + 1;

		for (int p = 1; p <= pmax; p++) {
			temp_index = (i - 1) * image_size + (j - 1);
			R1 = exp(p * 2.0 * ii * PI * pow(r[temp_index], 2.0));
			R2 = exp(p * Theta1[temp_index] * ii);

			p_index = p * (max + 1);
			s1 = s1 + (SR[p_index]) * (R1) + conj(SR[p_index]) * conj(R1)
					+ (SR[p]) * (R2) + conj(SR[p] * (R2));
			s2 = s2 + (SG[p_index]) * (R1) + conj(SG[p_index]) * conj(R1)
					+ (SG[p]) * (R2) + conj(SG[p] * (R2));
			s3 = s3 + (SB[p_index]) * (R1) + conj(SB[p_index]) * conj(R1)
					+ (SB[p]) * (R2) + conj(SB[p] * (R2));

			for (int q = 1; q <= qmax; q++) {

				R2 = exp(q * Theta1[temp_index] * ii);
				sum1 = sum1 + SR[p_index + q] * R1 * R2
						+ conj(SR[p_index + q] * R1 * R2)
						+ H1R[p_index + q] * conj(R1) * (R2)
						+ conj(H1R[p_index + q]) * R1 * conj(R2);
				sum2 = sum2 + SG[p_index + q] * R1 * R2
						+ conj(SG[p_index + q] * R1 * R2)
						+ H1G[p_index + q] * conj(R1) * (R2)
						+ conj(H1G[p_index + q]) * R1 * conj(R2);
				sum3 = sum3 + SB[p_index + q] * R1 * R2
						+ conj(SB[p_index + q] * R1 * R2)
						+ H1B[p_index + q] * conj(R1) * (R2)
						+ conj(H1B[p_index + q]) * R1 * conj(R2);
			}
		}

		h_3[temp_index] = (sum3 + s3 + SB[0]).real();
		h_2[temp_index] = (sum2 + s2 + SG[0]).real();
		h_1[temp_index] = (sum1 + s1 + SR[0]).real();

		sum1 = sum2 = sum3 = 0.0;
		s1 = s2 = s3 = 0;
	}
}

__global__ void QuaternionReconstructionKernel(double * r, double * Theta1,
		thrust::complex<double> *SR, thrust::complex<double> *SB,
		thrust::complex<double> *SG, thrust::complex<double> *H1R,
		thrust::complex<double> *H1B, thrust::complex<double> *H1G, double *h_1,
		double *h_2, double *h_3, int max, int image_size, int numOfThreads) {

	int threadID = blockIdx.x * blockDim.x + threadIdx.x;

//	__shared__ thrust::complex<double> sh_SR[(MAX + 1) * (MAX + 1)];
//	__shared__ thrust::complex<double> sh_SG[(MAX + 1) * (MAX + 1)];
//	__shared__ thrust::complex<double> sh_SB[(MAX + 1) * (MAX + 1)];
//
//	int shared_memory_load = ceil(1.0 * (MAX + 1) * (MAX + 1)) / blockDim.x;
//	if (shared_memory_load == 0)
//		shared_memory_load = 1;
//	int thread_memory_load_start = threadIdx.x * shared_memory_load;
//	int thread_memory_load_end = (threadIdx.x + 1) * shared_memory_load - 1;
//
//	if (thread_memory_load_start > (MAX + 1) * (MAX + 1) - 1)
//		thread_memory_load_start = (MAX + 1) * (MAX + 1) + 1;
//
//	if (thread_memory_load_end > (MAX + 1) * (MAX + 1) - 1)
//		thread_memory_load_end = (MAX + 1) * (MAX + 1) - 1;
//
//	if (threadIdx.x == (blockDim.x - 1)
//			&& thread_memory_load_end < ((MAX + 1) * (MAX + 1) - 1))
//		thread_memory_load_end = (MAX + 1) * (MAX + 1) - 1;
//
//	for (int index = thread_memory_load_start; index <= thread_memory_load_end;
//			index++) {
//		sh_SR[index] = SR[index];
//		sh_SG[index] = SG[index];
//		sh_SB[index] = SB[index];
//	}
//
//	__syncthreads();
//	if (threadID == 0) {
//		printf("load is : %i\n\n", shared_memory_load);
//		for (int i = 0; i < (MAX + 1) * (MAX + 1); i++)
//			if (SR[i].real() != sh_SR[i].real())
//				printf("i:%i>>(%f,%f) ", i, SR[i].real(), sh_SR[i].real());
//		printf("\n");
//	}

	int threadLoad = ceil((1.0 * image_size * image_size) / numOfThreads);
	if (threadLoad == 0)
		threadLoad = 1;

	int local_start = threadID * threadLoad;
	int local_end = (threadID + 1) * threadLoad - 1;

	if (local_start > image_size * image_size - 1)
		return;

	if (local_end > image_size * image_size - 1)
		local_end = image_size * image_size - 1;

	double PI = 3.1415926535897;
	thrust::complex<double> sum3 = 0.0, sum2 = 0.0, sum1 = 0.0, s1 = 0.0, s2 =
			0.0, s3 = 0.0;

	int pmax, qmax;
	pmax = qmax = max;
	thrust::complex<double> ii(0, 1);
	thrust::complex<double> R1, R2;

	int p_index = 0;
	int temp_index;
	for (int it = local_start; it <= local_end; it++) {

		int i = it / (image_size) + 1;
		int j = it % (image_size) + 1;
		for (int p = 1; p <= pmax; p++) {
			temp_index = (i - 1) * image_size + (j - 1);
			R1 = exp(p * 2.0 * ii * PI * pow(r[temp_index], 2.0));
			R2 = exp(p * Theta1[temp_index] * ii);

			p_index = p * (max + 1);
			s1 = s1 + (SR[p_index]) * (R1) + conj(SR[p_index]) * conj(R1)
					+ (SR[p]) * (R2) + conj(SR[p] * (R2));
			s2 = s2 + (SG[p_index]) * (R1) + conj(SG[p_index]) * conj(R1)
					+ (SG[p]) * (R2) + conj(SG[p] * (R2));
			s3 = s3 + (SB[p_index]) * (R1) + conj(SB[p_index]) * conj(R1)
					+ (SB[p]) * (R2) + conj(SB[p] * (R2));

			for (int q = 1; q <= qmax; q++) {

				R2 = exp(q * Theta1[temp_index] * ii);
				sum1 = sum1 + SR[p_index + q] * R1 * R2
						+ conj(SR[p_index + q] * R1 * R2)
						+ H1R[p_index + q] * conj(R1) * (R2)
						+ conj(H1R[p_index + q]) * R1 * conj(R2);
				sum2 = sum2 + SG[p_index + q] * R1 * R2
						+ conj(SG[p_index + q] * R1 * R2)
						+ H1G[p_index + q] * conj(R1) * (R2)
						+ conj(H1G[p_index + q]) * R1 * conj(R2);
				sum3 = sum3 + SB[p_index + q] * R1 * R2
						+ conj(SB[p_index + q] * R1 * R2)
						+ H1B[p_index + q] * conj(R1) * (R2)
						+ conj(H1B[p_index + q]) * R1 * conj(R2);
			}
		}
		h_3[temp_index] = (sum3 + s3 + SB[0]).real();
		h_2[temp_index] = (sum2 + s2 + SG[0]).real();
		h_1[temp_index] = (sum1 + s1 + SR[0]).real();

		sum1 = sum2 = sum3 = 0.0;
		s1 = s2 = s3 = 0;
	}
}
vector<double> interker(vector<double> x) {
	vector<double> hx(SIZE);
	double Absx[SIZE];
	for (unsigned i = 0; i < x.size(); i++) {
		Absx[i] = abs(x[i]);
	}

	for (unsigned i = 0; i < x.size(); i++) {
		if (Absx[i] <= 1)
			hx[i] = (3.0 / 2.0) * pow(Absx[i], 3)
					- (5.0 / 2.0) * pow(Absx[i], 2) + 1;
		else if (Absx[i] <= 2)
			hx[i] = (-1.0 / 2) * pow(Absx[i], 3) + (5.0 / 2.0) * pow(Absx[i], 2)
					- 4 * Absx[i] + 2;
		else
			hx[i] = 0;
	}
	return hx;
}

void errorComputing() {

	double x[SIZE];
	for (int i = 1; i <= SIZE; i++) {
		x[i - 1] = (2 * i - SIZE - 1) / (double) SIZE;
	}
	vector<vector<double> > r(SIZE, vector<double>(SIZE));

	for (int i = 0; i < SIZE; i++)
		for (int j = 0; j < SIZE; j++) {
			r[i][j] = sqrt(pow(x[i], 2) + pow(x[j], 2));
		}
	//%%%%%%%%%%NIRE Error%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	double sun1 = 0.0, sun11 = 0.0, sun22 = 0.0, sun222 = 0.0, sun2 = 0.0,
			sun3 = 0.0;
	int temp_index = 0;
	for (int i = 1; i <= SIZE; i++) {
		for (int j = 1; j <= SIZE; j++) {
			temp_index = (i - 1) * SIZE + (j - 1);
			if (r[i - 1][j - 1] <= 1) {
				sun2 = sun2 + pow(blue[i - 1][j - 1], 2.0);
				sun3 = sun3 + pow(h_3[temp_index] - blue[i - 1][j - 1], 2.0);
				// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				sun1 = sun1 + pow(red[i - 1][j - 1], 2.0);
				sun11 = sun11 + pow(h_1[temp_index] - red[i - 1][j - 1], 2.0);
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				sun22 = sun22 + pow(green[i - 1][j - 1], 2.0);
				sun222 = sun222
						+ pow(h_2[temp_index] - green[i - 1][j - 1], 2.0);
			} else {
				h_3[temp_index] = h_2[temp_index] = h_1[temp_index] = 255;
			}

		}
	}
	cout << endl;
	double dif3 = (sun3 / sun2);
	double dif2 = (sun222 / sun22);
	double dif1 = (sun11 / sun1);
	double dif = (dif2 + dif3 + dif1) / 3.0;
	cout << "Difference between original and reconstructed image = " << dif
			<< endl;
}

bool saveReconstructedImage(bool showImage = false) {
	int temp_index = 0;
	int size = SIZE;

	cv::Mat mat_red(size, size, CV_64FC1);
	cv::Mat mat_green(size, size, CV_64FC1);
	cv::Mat mat_blue(size, size, CV_64FC1);

	for (int i = 0; i < mat_red.rows; ++i)
		for (int j = 0; j < mat_red.cols; ++j) {
			temp_index = i * SIZE + j;
			mat_red.at<double>(i, j) = h_1[temp_index];
			mat_green.at<double>(i, j) = h_2[temp_index];
			mat_blue.at<double>(i, j) = h_3[temp_index];
		}

	Mat temp;
	vector<Mat> channels;
	channels.push_back(mat_blue);
	channels.push_back(mat_green);
	channels.push_back(mat_red);

	merge(channels, temp);
	Mat final_reconstructed_image;
	temp.convertTo(final_reconstructed_image, CV_8UC1);
	String imName = "reconstructed-" + std::to_string(MAX) + "-"
			+ imageFileName;
	imwrite(imName, final_reconstructed_image);
	if (showImage) {
		imshow(">>Reconst. Image", final_reconstructed_image);
		waitKey(0);
	}
	cout << "The reconstructed image is saved on disk with file name: "
			<< imName << endl;
	return true;
}

bool readImage(const String imageName) {

	ifstream image_file(imageName.c_str());

	if (!image_file.good()) {
		cerr << "Error: the image with name \"" << imageName
				<< "\" doesn't exist." << endl;
		return false;
	}
	cout << "Reading image: " << imageName << "......";
	Mat image = cv::imread(imageName);
	if (image.channels() != 3) {
		cerr << "Error: only images with three channels are supported" << endl;
		return false;
	}
	int tempRow = -1;
	int tempCol = -1;
	if (image.rows != SIZE || image.cols != SIZE) {
		tempRow = image.rows;
		tempCol = image.cols;
		resize(image, image, Size(SIZE, SIZE));
	}
	Mat bgr[3];   //destination array
	split(image, bgr);   //split source

	//Note: OpenCV uses BGR color order
	imwrite("blue.png", bgr[0]);   //blue channel
	imwrite("green.png", bgr[1]);   //green channel
	imwrite("red.png", bgr[2]);   //red channel

	vector<vector<int> > temp_blue(bgr[0].rows, vector<int>(bgr[0].cols));
	vector<vector<int> > temp_green(bgr[1].rows, vector<int>(bgr[1].cols));
	vector<vector<int> > temp_red(bgr[2].rows, vector<int>(bgr[2].cols));

	for (int i = 0; i < bgr[0].rows; i++)
		for (int j = 0; j < bgr[0].cols; j++) {
			if ((int) bgr[0].at<char>(i, j) >= 0) {
				temp_blue[i][j] = (int) bgr[0].at<char>(i, j);
			} else {
				temp_blue[i][j] = 256 + (int) bgr[0].at<char>(i, j);
			}
			if ((int) bgr[1].at<char>(i, j) >= 0) {
				temp_green[i][j] = (int) bgr[1].at<char>(i, j);
			} else {
				temp_green[i][j] = 256 + (int) bgr[1].at<char>(i, j);
			}

			if ((int) bgr[2].at<char>(i, j) >= 0) {
				temp_red[i][j] = (int) bgr[2].at<char>(i, j);
			} else {
				temp_red[i][j] = 256 + (int) bgr[2].at<char>(i, j);
			}
		}

	blue = temp_blue;
	green = temp_green;
	red = temp_red;

	cout << "Done!" << endl;
	if (tempCol != -1)
		cout << "The input image size is:" << tempRow << "x" << tempCol
				<< ". The image is resized to this dimensions: " << SIZE << "x"
				<< SIZE << "." << endl;
	return true;
}

static void show_usage(std::string name) {
	std::cerr << "Usage: " << name << " <option(s)>: \n"
					<< "\t--image\t set the input file image\n"
					<< "\t--threads\t set the used number of CPU threads.\n"
					<<"\t--gpu-threads\t set the number of threads per GPU block.\n"
					<<"\t--gpu-blocks\t set the number of GPU blocks.\n"
					<<"\t--num-replicas\t set the number of loop mitigate instances.\n"
					<<"\t--use-cpu-with-gpus\t 1 means use the CPU with the GPU(s), 0 otherwise.\n"
					<<"\t--num-GPUs\t set the number of used GPU(s)."
					<< std::endl;
}

bool setComputationalShare() {
	if (usingCPU)
		numOfDevices = numOfGPUs + 1;
	else {
		numOfDevices = numOfGPUs;
		int numOfshares = (SIZE / 4); // number of rings = size/2, and number of shared equals number of rings / 2
		int GPUShare = numOfshares / numOfGPUs;
		for (int k = 0; k < numOfDevices; k++)
			devicePowers.push_back(GPUShare);

		if (numOfGPUs * numOfshares < numOfGPUs)
			devicePowers[devicePowers.size() - 1] = numOfshares
					- (numOfGPUs * numOfshares);
		return true;
	}
//	devicePowers.push_back(860);
////	devicePowers.push_back(122);
////	devicePowers.push_back(122);
////	devicePowers.push_back(122);
//	devicePowers.push_back(164);

	int maxReplicas = 0;
	for (int index = 0; index < devicePowers.size(); index++)
		if (devicePowers[index] > maxReplicas)
			maxReplicas = devicePowers[index];
	numOfReplicas = numOfReplicasCPU = maxReplicas;
//	return true;

	int share;
	while (true) {
		cout << "\nPlease divide " << SIZE / 4 << " partitions over "
				<< numOfGPUs << " GPUs and " << numOfThreadsCPU
				<< " CPU threads, based on their computational power." << endl;
		for (int i = 0; i < numOfDevices; i++) {
			if (i < numOfDevices - 1) {
				cout << "GPU " << i << " share: ";
				cin >> share;
				while ((share < 0) || (share > SIZE / 4)) {
					cin.clear();
					cout << "Invalid entry, try again: GPU " << i << " share: ";
				}
				devicePowers.push_back(share);
			} else {
				cout << "CPU " << i << " share: ";
				cin >> share;
				while ((share < 0) || (share > SIZE / 4)) {
					cin.clear();
					cout << "Invalid entry, try again: CPU share: ";
				}
				devicePowers.push_back(share);
			}

		}
		int sum = 0;
		for (int k = 0; k < devicePowers.size(); k++)
			sum += devicePowers[k];
		if (sum == SIZE / 4) {
			int maxReplicas = 0;
			for (int index = 0; index < devicePowers.size(); index++)
				if (devicePowers[index] > maxReplicas)
					maxReplicas = devicePowers[index];
			numOfReplicas = numOfReplicasCPU = maxReplicas;
			return true;
		} else {
			cout << "The total sum of the entered share for the GPUs and CPU = "
					<< sum << " while it should be " << SIZE / 4 << endl;
			devicePowers.clear();
		}
	}
}
bool set_inputs(int argc, const char * argv[]) {
	if (argc < 7) {
		show_usage(argv[0]);
		return false;
	}

	for (int i = 1; i < argc; ++i) {
		std::string arg = argv[i];
		if (arg == "--image") {
			imageFileName = argv[i + 1];
			i++;
			if (!readImage(imageFileName)) {
				return false;
			}
		} else if (arg == "--threads") {
			numOfThreadsCPU = atoi(argv[i + 1]);
			i++;
			if (numOfThreadsCPU < 2) {
				std::cerr << "The number of CPU threads should be more than 1."
						<< endl;
				return false;
			}
			if (numOfThreadsCPU > (MAX + 1) * (MAX + 1)) {
				numOfThreadsCPU = (MAX + 1) * (MAX + 1);
				std::cerr << "The number of CPU threads is adjusted to "
						<< numOfThreadsCPU
						<< ". The maximum allowed number of CPU thread is (moment order +1)^2 = ("
						<< MAX << " + 1)^2." << endl;
			}
			cout << "The number of used CPU threads is set to: "
					<< numOfThreadsCPU << " threads." << endl;
		} else if (arg == "--gpu-threads") {
			noOfThreads = atoi(argv[i + 1]);
			i++;
			if ((noOfThreads % 32) != 0 || noOfThreads < 32) {
				std::cerr
						<< "The number of GPU threads per block should be a multiple of 32."
						<< endl;
				return false;
			}
			if (noOfThreads > 1024) {
				noOfThreads = 1024;
				std::cerr
						<< "The number of GPU threads per block is adjusted to "
						<< noOfThreads
						<< ". The maximum allowed number of GPU threads per block is 1024 thread."
						<< endl;
			}
			cout << "The number of GPU threads per block is set to: "
					<< noOfThreads << " threads." << endl;
		} else if (arg == "--gpu-blocks") {
			noOfBlocks = atoi(argv[i + 1]);

			cudaDeviceProp prop;
			cudaGetDeviceProperties(&prop, 0);
			int numOfMultiProcessors = prop.multiProcessorCount;
			i++;
			if ((noOfBlocks % numOfMultiProcessors) != 0
					|| noOfBlocks < numOfMultiProcessors) {
				std::cerr << "The number of GPU blocks should be a multiple of "
						<< numOfMultiProcessors
						<< ", as the number of SMs is:  "
						<< numOfMultiProcessors << endl;
				return false;
			}
			cout << "The number of GPU blocks is set to: " << noOfBlocks
					<< " threads. The number of GPU SMs is: "
					<< numOfMultiProcessors << endl;
		} else if (arg == "--num-replicas") {
			numOfReplicasCPU = atoi(argv[i + 1]);
			i++;
			if ((numOfReplicasCPU < 1) || numOfReplicasCPU > (SIZE / 4 - 1)) {
				std::cerr
						<< "The number of moment replicas for loop mitigation should be in the range [1,"
						<< (SIZE / 4 - 1) << "]." << endl;
				return false;
			}
			cout << "The number of moment  replicas is set to: "
					<< numOfReplicasCPU << "." << endl;
			numOfReplicasCPU++;
		} else if (arg == "--num-GPUs") {
			numOfGPUs = atoi(argv[i + 1]);
			i++;
			int nDevices;
			cudaGetDeviceCount(&nDevices);
			if ((numOfGPUs < 1) || numOfGPUs > nDevices) {
				std::cerr
						<< "The number of used GPUs should be in the range [1,"
						<< (nDevices) << "]." << endl;
				return false;
			}
			cout << "The number of used GPUs is set to: " << numOfGPUs << "."
					<< endl;
		} else if (arg == "--use-cpu-with-gpus") {
			usingCPU = atoi(argv[i + 1]);
			if (!((atoi(argv[i]) != 1) || (atoi(argv[i]) != 0))) {
				std::cerr
						<< "The --use-cpu-with-gpus should be assigned 1 or 0."
						<< endl;
				return false;
			}
			i++;
		} else {
			std::cerr << argv[i] << " is an invalid argument." << endl;
			return false;
		}

	}
	if (usingCPU)
		cout << "The computational resouces are " << numOfGPUs
				<< " GPU(s); each with " << noOfBlocks << " blocks and "
				<< noOfThreads << " thread per block. In addition to "
				<< numOfThreadsCPU << " CPU threads." << endl;
	else
		cout << "The computational resouces are " << numOfGPUs
				<< " GPU(s); each with " << noOfBlocks << " blocks and "
				<< noOfThreads << " thread per block." << endl;

	d_Ip.resize(numOfGPUs);
	d_Iq.resize(numOfGPUs);
	d_work.resize(numOfGPUs);
	d_sectorPerTrack.resize(numOfGPUs);
	d_accumSectors.resize(numOfGPUs);

	setComputationalShare();
	return true;
}

vector<DeviceWorkLoadInfo> divideKiOnDevices(int numOfDevices) {
	vector<DeviceWorkLoadInfo> deviceInfo(numOfDevices);

	cout << "Number of devices is :" << numOfDevices << endl;
	int s = 0;
	for (int i = 0; i < numOfDevices; i++)
		s += devicePowers[i];

	if (s != M / 2) {
		cerr
				<< "Error: Invalid devicePower: the sum should be (number of rings/2)= "
				<< M / 2 << " the sum is:" << s << endl;
		deviceInfo.clear();
		exit(-1);
		return deviceInfo;
	}

	int i = 1;
	int end = M;
	int count = 0;
	int deviceID = 0;
	int chunckSize = 0;
	while (count < M / 2) {
		chunckSize = devicePowers[deviceID];

		deviceInfo[deviceID].start1 = i;
		deviceInfo[deviceID].end1 = i + chunckSize - 1;

		i = deviceInfo[deviceID].end1 + 1;

		deviceInfo[deviceID].end2 = end;
		deviceInfo[deviceID].start2 = deviceInfo[deviceID].end2 - chunckSize
				+ 1;

		end = deviceInfo[deviceID].start2 - 1;
		count += chunckSize;
		deviceID++;
	}
//	for (int i = 0; i < numOfDevices; i++)
//		cout << deviceInfo[i].start1 << "," << deviceInfo[i].end1 << " "
//				<< deviceInfo[i].start2 << "," << deviceInfo[i].end2 << endl;
	return deviceInfo;
}

void computeIpIq() {
	struct timeval startAccesstime, endAccesstime;

	int Ki[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ki[i] = (2 * i + 1) * S;
	}

	double Ri[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ri[i] = (2 * i + 1) / (2.0 * M);
	}

	double deltaR = 1.0 / M;
	double U[M + 1];
	U[0] = Ri[0] - (deltaR / 2.0);

	for (int i = 1; i <= M; i++) {
		U[i] = Ri[i - 1] + deltaR / 2.0;
	}

	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		Theta1.push_back(x);
	}

	double PI = 3.1415926535897;

	for (int i = 1; i <= M; i++) {
		for (int n = 0; n < Ki[i - 1]; n++) {
			Theta1[i - 1][n] = (n + 0.5) * (2 * PI / Ki[i - 1]);
		}
	}

	//the upper and lower limit of the sectors angles
	double deltaTheta;

	vector<vector<double> > V;
	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		V.push_back(x);
	}

	for (int i = 0; i < M; i++) {
		deltaTheta = 2 * PI / Ki[i];
		V[i][0] = Theta1[i][0] - (deltaTheta / 2.0);
		for (int n = 1; n < Ki[i] + 1; n++) {
			V[i][n] = Theta1[i][n - 1] + (deltaTheta / 2.0);
		}
	}

	double r[M];
	for (int i = 0; i < M; i++)
		r[i] = Ri[i];

	gettimeofday(&startAccesstime, NULL);

	int pmax = MAX;
	int qmax = MAX;

	std::complex<double> I3(0, 1);
	std::complex<double> i1neg(0, -1);

#pragma omp parallel for private(I3) num_threads(numOfThreadsCPU)
	for (int p = 0; p <= pmax; p++) {
		for (int i = 1; i <= M; i++) {
			if (p == 0)
				I3 = 0.5 * (pow(U[i], 2.0) - pow(U[i - 1], 2.0));
			else
				I3 = (exp(p * 2.0 * i1neg * PI * pow(U[i], 2.0))
						- exp(p * 2.0 * i1neg * PI * pow(U[i - 1], 2.0)))
						/ (p * 4.0 * PI * i1neg);
			Ip[i - 1][p] = I3;
		}
	}

	std::complex<double> I(0, 0);
	std::complex<double> i1pos(0, 1);

	for (int i = 0; i < M; i++) {
		v2d x(MAX + 1, v1d(S * (2 * i + 1)));
		Iq.push_back(x);
	}

	//% %%%%%%%%%%%%%%%%%%%%%%I(q)%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#pragma omp parallel for private(I) num_threads(numOfThreadsCPU)
	for (int q = 0; q <= qmax; q++) {
		for (int i = 1; i <= M; i++) {
			for (int j = 1; j <= Ki[i - 1]; j++) {
				if (q == 0)
					I = V[i - 1][j] - V[i - 1][j - 1];
				else
					I = (i1pos / (double) q)
							* (exp(-1.0 * q * i1pos * V[i - 1][j])
									- exp(-1.0 * q * i1pos * V[i - 1][j - 1]));
				Iq[i - 1][q][j - 1] = I;
			}
		}
	}

	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "I_p & I_q computation time = " << totTime << " seconds." << endl;
}

bool transferImageIndependentDataToGPUs() {
	struct timeval startAccesstime, endAccesstime;
	gettimeofday(&startAccesstime, NULL);

	works = divideKiOnDevices(numOfDevices);
	if (works.size() == 0)
		return false;
	vector<vector<unsigned> > sectorPerTrack(numOfGPUs);
	vector<vector<unsigned> > accumSectors(numOfGPUs);

	numOfSectorsPerGPU.resize(numOfGPUs);

	int Ki[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ki[i] = (2 * i + 1) * S;
	}

	for (int g = 0; g < numOfGPUs; g++) {
		sectorPerTrack[g].resize(2 * (works[g].end1 - works[g].start1 + 1));
		accumSectors[g].resize(2 * (works[g].end1 - works[g].start1 + 1));
		numOfSectorsPerGPU[g] = 0;
		int k = 0;
		int acummSecs = 0;
		for (int i = works[g].start1; i <= works[g].end1; i++) {
			sectorPerTrack[g][k] = (2 * (i - 1) + 1) * S;
			accumSectors[g][k++] = acummSecs;
			acummSecs += (2 * (i - 1) + 1) * S;
			numOfSectorsPerGPU[g] += (2 * (i - 1) + 1) * S;
		}
		for (int i = works[g].start2; i <= works[g].end2; i++) {
			sectorPerTrack[g][k] = (2 * (i - 1) + 1) * S;
			accumSectors[g][k++] = acummSecs;
			acummSecs += (2 * (i - 1) + 1) * S;
			numOfSectorsPerGPU[g] += (2 * (i - 1) + 1) * S;
		}
	}

#pragma omp parallel  num_threads(numOfGPUs)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);
		int count = 0;
		int gpuWorkload = 2 * (works[gpuID].end1 - works[gpuID].start1 + 1);
		size_t a_size = sizeof(thrust::complex<double>)
				* size_t((MAX + 1) * gpuWorkload);
		cudaMalloc((void **) &d_Ip[gpuID], a_size);
		for (int i = works[gpuID].start1; i <= works[gpuID].end1;
				i++, count++) {
			cudaMemcpyAsync(d_Ip[gpuID] + count * ((MAX + 1)), &Ip[i - 1][0],
					sizeof(thrust::complex<double>) * (MAX + 1),
					cudaMemcpyHostToDevice);
		}
		for (int i = works[gpuID].start2; i <= works[gpuID].end2;
				i++, count++) {
			cudaMemcpyAsync(d_Ip[gpuID] + count * ((MAX + 1)), &Ip[i - 1][0],
					sizeof(thrust::complex<double>) * (MAX + 1),
					cudaMemcpyHostToDevice);
		}

		count = 0;
		a_size = sizeof(thrust::complex<double>)
				* size_t((MAX + 1) * numOfSectorsPerGPU[gpuID]);

		cudaMalloc((void **) &d_Iq[gpuID], a_size);
		int currentPosistion = 0;
		for (int i = works[gpuID].start1; i <= works[gpuID].end1; i++) {
			for (int q = 0; q < (MAX + 1); q++) {
				cudaMemcpyAsync(d_Iq[gpuID] + currentPosistion,
						&Iq[i - 1][q][0],
						sizeof(thrust::complex<double>) * Ki[i - 1],
						cudaMemcpyHostToDevice);
				currentPosistion += Ki[i - 1];
			}
		}

		for (int i = works[gpuID].start2; i <= works[gpuID].end2; i++) {
			for (int q = 0; q < (MAX + 1); q++) {
				cudaMemcpyAsync(d_Iq[gpuID] + currentPosistion,
						&Iq[i - 1][q][0],
						sizeof(thrust::complex<double>) * Ki[i - 1],
						cudaMemcpyHostToDevice);
				currentPosistion += Ki[i - 1];
			}
		}

		a_size = sizeof(unsigned) * size_t(sectorPerTrack[gpuID].size());
		cudaMalloc((void **) &d_sectorPerTrack[gpuID], a_size);
		cudaMemcpyAsync(d_sectorPerTrack[gpuID], &sectorPerTrack[gpuID][0],
				a_size, cudaMemcpyHostToDevice);

		a_size = sizeof(unsigned) * size_t(accumSectors[gpuID].size());
		cudaMalloc((void **) &d_accumSectors[gpuID], a_size);
		cudaMemcpyAsync(d_accumSectors[gpuID], &accumSectors[gpuID][0], a_size,
				cudaMemcpyHostToDevice);

		if (cudaMalloc((void **) &d_work[gpuID], sizeof(DeviceWorkLoadInfo))
				!= cudaSuccess)
			printf("FAILED TO ALLOCATE MATRIX ON DEVICE!\n");
		cudaMemcpyAsync(d_work[gpuID], &works[gpuID],
				sizeof(DeviceWorkLoadInfo), cudaMemcpyHostToDevice);
	}

#pragma omp parallel  num_threads(numOfGPUs)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);
		cudaDeviceSynchronize();
	}
	gettimeofday(&endAccesstime, NULL);

	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image independent data transfer HtoD time = " << totTime
			<< " seconds. # of GPUs: " << numOfGPUs << endl << endl;

	return true;
}

bool Quaternion_moment_computation_MultiGPU_CPU() {
	struct timeval startAccesstime, endAccesstime;
	struct timeval startAccesstime1, endAccesstime1;

	int Ki[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ki[i] = (2 * i + 1) * S;
	}

	double Ri[SIZE / 2];
	for (int i = 0; i < M; i++) {
		Ri[i] = (2 * i + 1) / (2.0 * M);
	}

	double PI = 3.1415926535897;
	double deltaTheta;

	double r[M];
	for (int i = 0; i < M; i++)
		r[i] = Ri[i];

	int s[SIZE];
	for (int i = 0; i < SIZE; i++)
		s[i] = i + 1;

	double xi[SIZE];
	for (int i = 0; i < SIZE; i++)
		xi[i] = (2.0 * s[i] - SIZE - 1) / SIZE;

	double yi[SIZE];
	for (int i = 0; i < SIZE; i++)
		yi[i] = (2.0 * s[i] - SIZE - 1) / SIZE;

	double del = 2.0 / SIZE;   //% the pixel width
	const int N = SIZE;

	gettimeofday(&startAccesstime, NULL);

	for (int i = 0; i < M; i++) {
		vector<double> x(S * (2 * i + 1));
		std::fill(x.begin(), x.end(), 0);
		fR.push_back(x);
		fB.push_back(x);
		fG.push_back(x);
	}

#pragma omp parallel for schedule(dynamic,1) num_threads(numOfThreadsCPU)
	for (int i = 1; i <= M; i++) {
		for (int j = 1; j <= Ki[i - 1]; j++) {

			int k1 = ceil(r[i - 1] * cos(Theta1[i - 1][j - 1]) / del) + N / 2;
			int k2 = ceil(r[i - 1] * sin(Theta1[i - 1][j - 1]) / del) + N / 2;

			int lSize = (k1 + 2) - (k1 - 1) + 1;
			int mSize = (k2 + 2) - (k2 - 1) + 1;
			std::vector<int> l(lSize);
			std::vector<int> m(mSize);
			std::vector<int> inx(lSize);
			std::vector<int> iny(mSize);

			for (int i = 0, count = k1 - 1; i < lSize; i++) {
				l[i] = count++;
				inx[i] = l[i];
			}

			for (int i = 0, count = k2 - 1; i < mSize; i++) {
				m[i] = count++;
				iny[i] = m[i];
			}

			// construct image matrix
			for (int i = 0; i < lSize; i++) {
				if (l[i] > N) {
					inx[i] = l[i] % N;
					l[i] = N - (l[i] % N) + 1;
					continue;
				}
				if (l[i] < 1) {
					inx[i] = N - abs(l[i]);
					l[i] = N - ((l[i] + N) - 1);
				}

			}

			for (int i = 0; i < mSize; i++) {
				if (m[i] > N) {
					iny[i] = m[i] % N;
					m[i] = N - (m[i] % N) + 1;
					continue;
				}
				if (m[i] < 1) {
					iny[i] = N - abs(m[i]);
					m[i] = N - ((m[i] + N) - 1);
				}

			}
			//%construct the image 4*4 matrix
			vector<vector<double> > Imatrix1(lSize, vector<double>(mSize));
			vector<vector<double> > Imatrix2(lSize, vector<double>(mSize));
			vector<vector<double> > Imatrix3(lSize, vector<double>(mSize));

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix1[i][j] = red[l[i] - 1][m[j] - 1];
				}
			}

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix2[i][j] = green[l[i] - 1][m[j] - 1];
				}
			}

			for (int i = 0; i < lSize; i++) {
				for (int j = 0; j < mSize; j++) {
					Imatrix3[i][j] = blue[l[i] - 1][m[j] - 1];
				}
			}
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//% construct kernel
			std::vector<double> xl(lSize);

			for (int i = 0; i < lSize; i++)
				xl[i] = xi[inx[i] - 1];

			//			double * ym = new double[mSize];
			std::vector<double> ym(mSize);

			for (int i = 0; i < mSize; i++)
				ym[i] = yi[iny[i] - 1];

			// % compute the kernel function using 'interker' function
			std::vector<double> tempXl(lSize);

			for (int ww = 0; ww < lSize; ww++) {
				tempXl[ww] = (r[i - 1] * cos(Theta1[i - 1][j - 1]) - xl[ww])
						/ del;
			}

			std::vector<double> h1(lSize);
			vector<double> hx = interker(tempXl);  //% for x
			for (int i = 0; i < lSize; i++)
				h1[i] = hx[i];

			std::vector<double> tempYm(mSize);
			for (int ww = 0; ww < mSize; ww++)
				tempYm[ww] = (r[i - 1] * sin(Theta1[i - 1][j - 1]) - ym[ww])
						/ del;

			std::vector<double> h2(mSize);
			hx = interker(tempYm);  //% for x
			for (int i = 0; i < lSize; i++)
				h2[i] = hx[i];

			vector<vector<double> > hh(lSize, vector<double>(mSize));
			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++)
					hh[i][j] = h1[i] * h2[j];

			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//% 2D convolution between the kernel and the image matrix
			vector<vector<double> > P1(lSize, vector<double>(mSize));
			vector<vector<double> > P2(lSize, vector<double>(mSize));
			vector<vector<double> > P3(lSize, vector<double>(mSize));

			double sum1 = 0, sum2 = 0, sum3 = 0;
			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P1[i][j] = Imatrix1[i][j] * hh[i][j];
					sum1 += P1[i][j];
				}

			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P2[i][j] = Imatrix2[i][j] * hh[i][j];
					sum2 += P2[i][j];
				}

			for (int i = 0; i < lSize; i++)
				for (int j = 0; j < mSize; j++) {
					P3[i][j] = Imatrix3[i][j] * hh[i][j];
					sum3 += P3[i][j];
				}
			fR[i - 1][j - 1] = sum1;
			fG[i - 1][j - 1] = sum2;
			fB[i - 1][j - 1] = sum3;

			//clearing area
			P1.clear();
			P1.clear();
			P3.clear();
			hh.clear();
			h2.clear();
			tempYm.clear();
			h1.clear();
			tempXl.clear();
			ym.clear();
			xl.clear();
			Imatrix3.clear();
			Imatrix2.clear();
			Imatrix1.clear();
			iny.clear();
			inx.clear();
			m.clear();
			l.clear();

		}
	}
	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Interpolation time = " << totTime << " seconds." << endl;

	//multiGPU
	vector<double *> d_fR(numOfGPUs);
	vector<double *> d_fB(numOfGPUs);
	vector<double *> d_fG(numOfGPUs);

	vector<thrust::complex<double> *> d_SG(numOfGPUs);
	vector<thrust::complex<double> *> d_SB(numOfGPUs);
	vector<thrust::complex<double> *> d_SR(numOfGPUs);
	vector<thrust::complex<double> *> d_H1G(numOfGPUs);
	vector<thrust::complex<double> *> d_H1B(numOfGPUs);
	vector<thrust::complex<double> *> d_H1R(numOfGPUs);

	gettimeofday(&startAccesstime, NULL);

	omp_set_num_threads(numOfGPUs);
	cout << "\n----------------------------------------" << endl;
	cout << "Quat moment computation started on GPU...." << endl;
	cout
			<< "CPU threads are set to the number of GPUs for memory transfer. Each thread transfer GPU data concurrently."
			<< endl;
#pragma omp parallel  num_threads(numOfGPUs)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);
		int count = 0;
		size_t a_size = sizeof(double) * size_t(numOfSectorsPerGPU[gpuID]);

		cudaMalloc((void **) &d_fR[gpuID], a_size);
		cudaMalloc((void **) &d_fB[gpuID], a_size);
		cudaMalloc((void **) &d_fG[gpuID], a_size);

		int currentPosition = 0;
		for (int i = works[gpuID].start1; i <= works[gpuID].end1; i++) {
			cudaMemcpyAsync(d_fR[gpuID] + currentPosition, &fR[i - 1][0],
					sizeof(double) * ((2 * (i - 1) + 1) * S),
					cudaMemcpyHostToDevice);
			cudaMemcpyAsync(d_fG[gpuID] + currentPosition, &fG[i - 1][0],
					sizeof(double) * ((2 * (i - 1) + 1) * S),
					cudaMemcpyHostToDevice);
			cudaMemcpyAsync(d_fB[gpuID] + currentPosition, &fB[i - 1][0],
					sizeof(double) * ((2 * (i - 1) + 1) * S),
					cudaMemcpyHostToDevice);
			currentPosition += ((2 * (i - 1) + 1) * S);
		}

		for (int i = works[gpuID].start2; i <= works[gpuID].end2; i++) {
			cudaMemcpyAsync(d_fR[gpuID] + currentPosition, &fR[i - 1][0],
					sizeof(double) * ((2 * (i - 1) + 1) * S),
					cudaMemcpyHostToDevice);
			cudaMemcpyAsync(d_fG[gpuID] + currentPosition, &fG[i - 1][0],
					sizeof(double) * ((2 * (i - 1) + 1) * S),
					cudaMemcpyHostToDevice);
			cudaMemcpyAsync(d_fB[gpuID] + currentPosition, &fB[i - 1][0],
					sizeof(double) * ((2 * (i - 1) + 1) * S),
					cudaMemcpyHostToDevice);
			currentPosition += ((2 * (i - 1) + 1) * S);
		}

		a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * (MAX + 1))
				* numOfReplicasCPU;

		cudaMalloc((void **) &d_SG[gpuID], a_size);
		cudaMemset(d_SG[gpuID], 0, a_size);

		cudaMalloc((void **) &d_SB[gpuID], a_size);
		cudaMemset(d_SB[gpuID], 0, a_size);

		cudaMalloc((void **) &d_SR[gpuID], a_size);
		cudaMemset(d_SR[gpuID], 0, a_size);

		cudaMalloc((void **) &d_H1G[gpuID], a_size);
		cudaMemset(d_H1G[gpuID], 0, a_size);

		cudaMalloc((void **) &d_H1B[gpuID], a_size);
		cudaMemset(d_H1B[gpuID], 0, a_size);

		cudaMalloc((void **) &d_H1R[gpuID], a_size);
		cudaMemset(d_H1R[gpuID], 0, a_size);

	}

#pragma omp parallel  num_threads(numOfGPUs)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);
		cudaDeviceSynchronize();
	}

	gettimeofday(&endAccesstime, NULL);

	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Quater. moment multi-GPU: data transfer HtoD time = " << totTime
			<< " seconds. # of GPUs: " << numOfGPUs << endl;

	omp_set_num_threads(numOfGPUs);
	gettimeofday(&startAccesstime, NULL);
	//Multi(GPU + CPU)
#pragma omp parallel  num_threads(numOfGPUs)
	//for(int gpuID = 0 ; gpuID <numOfGPUs ; gpuID++)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);
		QuaternionMomentKernelMulti<<<noOfBlocks, noOfThreads>>>(d_Ip[gpuID],
				d_Iq[gpuID], d_fR[gpuID], d_fB[gpuID], d_fG[gpuID], d_SR[gpuID],
				d_SB[gpuID], d_SG[gpuID], d_H1R[gpuID], d_H1B[gpuID],
				d_H1G[gpuID], M, MAX, totNoOfThreads, d_work[gpuID],
				d_sectorPerTrack[gpuID], d_accumSectors[gpuID],
				numOfReplicasCPU);
	}
	//CPU part starts
	size_t a_size = ((MAX + 1) * (MAX + 1)) * sizeof(thrust::complex<double>);
	thrust::complex<double> * h_tempSR = (thrust::complex<double>*) malloc(
			a_size);
	thrust::complex<double> * h_tempSB = (thrust::complex<double>*) malloc(
			a_size);
	thrust::complex<double> * h_tempSG = (thrust::complex<double>*) malloc(
			a_size);

	thrust::complex<double> * h_tempH1R = (thrust::complex<double>*) malloc(
			a_size);
	thrust::complex<double> * h_tempH1B = (thrust::complex<double>*) malloc(
			a_size);
	thrust::complex<double> * h_tempH1G = (thrust::complex<double>*) malloc(
			a_size);

	int loopSize = MAX + 1;
	int size = loopSize * loopSize / numOfThreadsCPU;
	DeviceWorkLoadInfo workLoad = works[numOfGPUs];

	omp_set_num_threads(numOfThreadsCPU);
	double pii = 7.0 / 22.0;

	gettimeofday(&startAccesstime1, NULL);

#pragma omp parallel  num_threads(numOfThreadsCPU)
	{
		int threadID = omp_get_thread_num();

		int local_start = threadID * size;
		int local_end = (threadID + 1) * size - 1;
		if (threadID == numOfThreadsCPU - 1)
			local_end = loopSize * loopSize - 1;

		int p, q;
		std::complex<double> S1 = 0, S2 = 0, S3 = 0, S4 = 0, S5 = 0, S6 = 0;
		std::complex<double> YR = 0, YG = 0, YB = 0;
		int z1, z2, z3;

		for (int i = local_start; i <= local_end; i++) {
			p = i / (loopSize);
			q = i % loopSize;

			S1 = S2 = S3 = S4 = S5 = S6 = 0;
			for (int i = workLoad.start1; i <= workLoad.end2; i++) {
				for (int j = 1; j <= 2 * i - 1; j++) {
					z1 = 4 * i - 1 - j;
					z2 = 4 * i - 2 + j;
					z3 = 8 * i - 3 - j;

					YR = fR[i - 1][j - 1] * Iq[i - 1][q][j - 1]
							+ fR[i - 1][z1 - 1] * Iq[i - 1][q][z1 - 1]
							+ fR[i - 1][z2 - 1] * Iq[i - 1][q][z2 - 1]
							+ fR[i - 1][z3 - 1] * Iq[i - 1][q][z3 - 1];
					YG = fG[i - 1][j - 1] * Iq[i - 1][q][j - 1]
							+ fG[i - 1][z1 - 1] * Iq[i - 1][q][z1 - 1]
							+ fG[i - 1][z2 - 1] * Iq[i - 1][q][z2 - 1]
							+ fG[i - 1][z3 - 1] * Iq[i - 1][q][z3 - 1];
					YB = fB[i - 1][j - 1] * Iq[i - 1][q][j - 1]
							+ fB[i - 1][z1 - 1] * Iq[i - 1][q][z1 - 1]
							+ fB[i - 1][z2 - 1] * Iq[i - 1][q][z2 - 1]
							+ fB[i - 1][z3 - 1] * Iq[i - 1][q][z3 - 1];

					S1 = S1 + (YR) * Ip[i - 1][p];
					S2 = S2 + (YG) * Ip[i - 1][p];
					S3 = S3 + (YB) * Ip[i - 1][p];

					S4 = S4 + (YR) * conj(Ip[i - 1][p]);
					S5 = S5 + (YG) * conj(Ip[i - 1][p]);
					S6 = S6 + (YB) * conj(Ip[i - 1][p]);
				}
			}

			h_tempSR[p * loopSize + q] = pii * S1;
			h_tempSG[p * loopSize + q] = pii * S2;
			h_tempSB[p * loopSize + q] = pii * S3;

			h_tempH1R[p * loopSize + q] = pii * S4;
			h_tempH1G[p * loopSize + q] = pii * S5;
			h_tempH1B[p * loopSize + q] = pii * S6;
		}
	}
	//CPU part ends

	gettimeofday(&endAccesstime1, NULL);

	totTime = (double) ((endAccesstime1.tv_usec - startAccesstime1.tv_usec)
			/ 1.0e6 + endAccesstime1.tv_sec - startAccesstime1.tv_sec);
	cout << "partial cpu time = " << totTime << endl;

	cudaError err;

	omp_set_num_threads(numOfGPUs);
#pragma omp parallel  num_threads(numOfGPUs)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);
		gpuErrchk(cudaPeekAtLastError());
		gpuErrchk(cudaDeviceSynchronize());
		cudaError err = cudaGetLastError();
		if (cudaSuccess != err) {
			cout << cudaGetErrorString(err) << endl;
			exit(-1);
		}
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Quater moment computation on multi-GPU computation time = "
			<< totTime << " seconds. # of threads: " << totNoOfThreads
			<< " per a GPU, # of GPUs: " << numOfGPUs;
	if (usingCPU)
		cout << " and # of CPU threads: " << numOfThreadsCPU;
	cout << endl;

	gettimeofday(&startAccesstime, NULL);

	//adding CPU
	a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * (MAX + 1))
			* numOfReplicasCPU;

	h_SB = (thrust::complex<double>*) malloc(a_size);
	h_SG = (thrust::complex<double>*) malloc(a_size);
	h_SR = (thrust::complex<double>*) malloc(a_size);

	h_H1B = (thrust::complex<double>*) malloc(a_size);
	h_H1G = (thrust::complex<double>*) malloc(a_size);
	h_H1R = (thrust::complex<double>*) malloc(a_size);

	for (int gpuID = 0; gpuID < numOfGPUs; gpuID++) {
		cudaSetDevice(gpuID);

		cudaMemcpy(h_SB, d_SB[gpuID], a_size, cudaMemcpyDeviceToHost);
		cudaMemcpy(h_SG, d_SG[gpuID], a_size, cudaMemcpyDeviceToHost);
		cudaMemcpy(h_SR, d_SR[gpuID], a_size, cudaMemcpyDeviceToHost);

		cudaMemcpy(h_H1B, d_H1B[gpuID], a_size, cudaMemcpyDeviceToHost);
		cudaMemcpy(h_H1G, d_H1G[gpuID], a_size, cudaMemcpyDeviceToHost);
		cudaMemcpy(h_H1R, d_H1R[gpuID], a_size, cudaMemcpyDeviceToHost);

		cudaDeviceSynchronize();

		int index1 = 0, index2 = 0;
		for (int i = 0; i < (MAX + 1); i++)
			for (int j = 0; j < (MAX + 1); j++) {
				index1 = i * (MAX + 1) + j;
				for (int partitionID = 0; partitionID < numOfReplicasCPU;
						partitionID++) {
					index2 = partitionID * (MAX + 1) * (MAX + 1) + index1;
					h_tempSR[index1] += h_SR[index2];
					h_tempSG[index1] += h_SG[index2];
					h_tempSB[index1] += h_SB[index2];
					h_tempH1R[index1] += h_H1R[index2];
					h_tempH1G[index1] += h_H1G[index2];
					h_tempH1B[index1] += h_H1B[index2];
				}
			}

		//adding up gpu[gpuID] results
		for (int i = 0; i < ((MAX + 1) * (MAX + 1)); i++) {
			h_SR[i] = h_tempSR[i];
			h_SG[i] = h_tempSG[i];
			h_SB[i] = h_tempSB[i];
			h_H1R[i] = h_tempH1R[i];
			h_H1G[i] = h_tempH1G[i];
			h_H1B[i] = h_tempH1B[i];
		}
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Quater. moment multi-GPU: data transfer DtoH time = " << totTime
			<< " seconds." << endl;
	cout << "Quaternion moment is copied from devices to host." << endl;

	for (int gpuID = 0; gpuID < numOfGPUs; gpuID++) {
		cudaSetDevice(gpuID);
		cudaFree(d_Ip[gpuID]);
		cudaFree(d_Iq[gpuID]);
		cudaFree(d_fB[gpuID]);
		cudaFree(d_fG[gpuID]);
		cudaFree(d_fR[gpuID]);
		cudaFree(d_SB[gpuID]);
		cudaFree(d_SG[gpuID]);
		cudaFree(d_SR[gpuID]);
		cudaFree(d_H1B[gpuID]);
		cudaFree(d_H1G[gpuID]);
		cudaFree(d_H1R[gpuID]);
	}

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << "GPU memory de-allocation: " << cudaGetErrorString(err) << endl;
		exit(-1);
	}
	return true;
}

void GPU_Image_Reconstruct_Computation() {

	double x[SIZE];

	for (int i = 1; i <= SIZE; i++) {
		x[i - 1] = ((2 * i) - SIZE - 1) / (double) SIZE;
	}

	vector<vector<double> > Theta1(SIZE, vector<double>(SIZE));
	vector<vector<double> > r(SIZE, vector<double>(SIZE));

	for (int i = 0; i < SIZE; i++)
		for (int j = 0; j < SIZE; j++) {
			r[i][j] = sqrt(pow(x[i], 2) + pow(x[j], 2));
			Theta1[i][j] = std::atan2(x[j], x[i]);
		}

	struct timeval startAccesstime, endAccesstime;

	int numOfThreads = noOfBlocks * noOfThreads;
	cout << "\n----------------------------------------" << endl;
	cout << "Quat image reconstruction computation started on GPU...." << endl;

	gettimeofday(&startAccesstime, NULL);

//r
	double * d_r;
	size_t a_size = sizeof(double) * size_t(SIZE * SIZE);
	cudaMalloc((void **) &d_r, a_size);
//theta1
	double * d_Theta1;
	a_size = sizeof(double) * size_t(SIZE * SIZE);
	cudaMalloc((void **) &d_Theta1, a_size);

	for (int i = 0, currentPosition = 0; i < SIZE; i++) {
		cudaMemcpyAsync(d_r + currentPosition, &r[i][0], sizeof(double) * SIZE,
				cudaMemcpyHostToDevice);
		cudaMemcpyAsync(d_Theta1 + currentPosition, &Theta1[i][0],
				sizeof(double) * SIZE, cudaMemcpyHostToDevice);
		currentPosition += SIZE;
	}

	a_size = sizeof(thrust::complex<double>) * size_t((MAX + 1) * (MAX + 1));

	thrust::complex<double> * d_SG;
	cudaMalloc((void **) &d_SG, a_size);
	cudaMemcpy(d_SG, h_SG, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_SB;
	cudaMalloc((void **) &d_SB, a_size);
	cudaMemcpy(d_SB, h_SB, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_SR;
	cudaMalloc((void **) &d_SR, a_size);
	cudaMemcpy(d_SR, h_SR, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_H1G;
	cudaMalloc((void **) &d_H1G, a_size);
	cudaMemcpy(d_H1G, h_H1G, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_H1B;
	cudaMalloc((void **) &d_H1B, a_size);
	cudaMemcpy(d_H1B, h_H1B, a_size, cudaMemcpyHostToDevice);

	thrust::complex<double> * d_H1R;
	cudaMalloc((void **) &d_H1R, a_size);
	cudaMemcpy(d_H1R, h_H1R, a_size, cudaMemcpyHostToDevice);

	a_size = sizeof(double) * size_t(SIZE * SIZE);
	double * d_h1;
	cudaMalloc((void **) &d_h1, a_size);

	double * d_h2;
	cudaMalloc((void **) &d_h2, a_size);

	double * d_h3;
	cudaMalloc((void **) &d_h3, a_size);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	cudaError err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << "GPU memory allocation: " << cudaGetErrorString(err) << endl;
		exit(-1);
	}

	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Reconst on GPU: data transfer HtoD time  = " << totTime
			<< endl;

	gettimeofday(&startAccesstime, NULL);
	QuaternionReconstructionKernel<<<noOfBlocks, noOfThreads>>>(d_r, d_Theta1,
			d_SR, d_SB, d_SG, d_H1R, d_H1B, d_H1G, d_h1, d_h2, d_h3,
			MAX, SIZE, numOfThreads);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	err = cudaGetLastError();
	if (cudaSuccess != err) {
		cout << "Reconst kernel: " << cudaGetErrorString(err) << endl;
		exit(-1);
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image reconst on a single GPU running time  = " << totTime
			<< " seconds." << " # of GPU threads: " << totNoOfThreads << endl;

	gettimeofday(&startAccesstime, NULL);

	a_size = sizeof(double) * size_t(SIZE * SIZE);
	h_1 = (double*) malloc(a_size);
	h_2 = (double*) malloc(a_size);
	h_3 = (double*) malloc(a_size);

	cudaMemcpy(h_1, d_h1, a_size, cudaMemcpyDeviceToHost);
	cudaMemcpy(h_2, d_h2, a_size, cudaMemcpyDeviceToHost);
	cudaMemcpy(h_3, d_h3, a_size, cudaMemcpyDeviceToHost);

	gpuErrchk(cudaPeekAtLastError());
	gpuErrchk(cudaDeviceSynchronize());
	err = cudaGetLastError();

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Reconst on GPU: data transfer DtoH time  = " << totTime
			<< " seconds." << endl;

	cudaFree(d_SB);
	cudaFree(d_SG);
	cudaFree(d_SR);
	cudaFree(d_H1B);
	cudaFree(d_H1G);
	cudaFree(d_H1R);
}

void GPU_CPU_Reconstruct(int GPUs, int singleGPUShare, int CPUShare) {

	cout << "\n----------------------------------------" << endl;
	double x[SIZE];
	for (int i = 1; i <= SIZE; i++) {
		x[i - 1] = ((2 * i) - SIZE - 1) / (double) SIZE;
	}

	vector<vector<double> > Theta1(SIZE, vector<double>(SIZE));
	vector<vector<double> > r(SIZE, vector<double>(SIZE));
	for (int i = 0; i < SIZE; i++)
		for (int j = 0; j < SIZE; j++) {
			r[i][j] = sqrt(pow(x[i], 2) + pow(x[j], 2));
			Theta1[i][j] = std::atan2(x[j], x[i]);
		}

	struct timeval startAccesstime, endAccesstime;

	thrust::complex<double> sum3 = 0.0, sum2 = 0.0, sum1 = 0.0;
	thrust::complex<double> s1 = 0.0, s2 = 0.0, s3 = 0.0;
	thrust::complex<double> ii(0, 1);

	int pmax, qmax;
	pmax = qmax = MAX;

	int numOfThreads = noOfBlocks * noOfThreads;
	unsigned totNumOfIterations = SIZE * SIZE;

	vector<ThreadInfo> shares(GPUs + 1);

	int totalShares = GPUs * singleGPUShare + CPUShare;	//1 here for CPU Share
	int NumOfIterationsPerGPU = ceil(
			(1.0 * singleGPUShare) / totalShares * totNumOfIterations);

	int start = 0;
	for (int GPUID = 0; GPUID < GPUs; GPUID++) {
		shares[GPUID].i = start;
		start += NumOfIterationsPerGPU;
		shares[GPUID].j = start - 1;
		cout << "GPU " << GPUID << " start: " << shares[GPUID].i << " end: "
				<< shares[GPUID].j << endl;
	}

	if (!usingCPU) {

	} else {
		shares[shares.size() - 1].i = start;
		shares[shares.size() - 1].j = totNumOfIterations - 1;
	}
	cout << "cpu  start: " << shares[shares.size() - 1].i << " end: "
			<< shares[shares.size() - 1].j << endl;
	//load balance end

	gettimeofday(&startAccesstime, NULL);

	vector<double *> d_r(GPUs);
	vector<double *> d_Theta1(GPUs);
	vector<thrust::complex<double> *> d_SG(GPUs);
	vector<thrust::complex<double> *> d_SB(GPUs);
	vector<thrust::complex<double> *> d_SR(GPUs);
	vector<thrust::complex<double> *> d_H1G(GPUs);
	vector<thrust::complex<double> *> d_H1B(GPUs);
	vector<thrust::complex<double> *> d_H1R(GPUs);

	vector<double *> d_h1(GPUs);
	vector<double *> d_h2(GPUs);
	vector<double *> d_h3(GPUs);
	cudaError err;

	omp_set_num_threads(GPUs);
#pragma omp parallel num_threads(GPUs)
	//for (int gpuID = 0; gpuID < GPUs; gpuID++)
	{

		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);

		size_t a_size = sizeof(double) * size_t(SIZE * SIZE);
		cudaMalloc((void **) &d_r[gpuID], a_size);

		//theta1
		a_size = sizeof(double) * size_t(SIZE * SIZE);
		cudaMalloc((void **) &d_Theta1[gpuID], a_size);

		for (int i = 0, currentPosition = 0; i < SIZE; i++) {
			cudaMemcpyAsync(d_r[gpuID] + currentPosition, &r[i][0],
					sizeof(double) * SIZE, cudaMemcpyHostToDevice);
			cudaMemcpyAsync(d_Theta1[gpuID] + currentPosition, &Theta1[i][0],
					sizeof(double) * SIZE, cudaMemcpyHostToDevice);
			currentPosition += SIZE;
		}

		a_size = sizeof(thrust::complex<double>)
				* size_t((MAX + 1) * (MAX + 1));

		cudaMalloc((void **) &d_SG[gpuID], a_size);
		cudaMemcpy(d_SG[gpuID], h_SG, a_size, cudaMemcpyHostToDevice);

		cudaMalloc((void **) &d_SB[gpuID], a_size);
		cudaMemcpy(d_SB[gpuID], h_SB, a_size, cudaMemcpyHostToDevice);

		cudaMalloc((void **) &d_SR[gpuID], a_size);
		cudaMemcpy(d_SR[gpuID], h_SR, a_size, cudaMemcpyHostToDevice);

		cudaMalloc((void **) &d_H1G[gpuID], a_size);
		cudaMemcpy(d_H1G[gpuID], h_H1G, a_size, cudaMemcpyHostToDevice);

		cudaMalloc((void **) &d_H1B[gpuID], a_size);
		cudaMemcpy(d_H1B[gpuID], h_H1B, a_size, cudaMemcpyHostToDevice);

		cudaMalloc((void **) &d_H1R[gpuID], a_size);
		cudaMemcpy(d_H1R[gpuID], h_H1R, a_size, cudaMemcpyHostToDevice);

		a_size = sizeof(double) * size_t(SIZE * SIZE);

		cudaMalloc((void **) &d_h1[gpuID], a_size);
		cudaMemset(d_h1[gpuID], 0, a_size);

		cudaMalloc((void **) &d_h2[gpuID], a_size);
		cudaMemset(d_h2[gpuID], 0, a_size);

		cudaMalloc((void **) &d_h3[gpuID], a_size);
		cudaMemset(d_h3[gpuID], 0, a_size);

		gpuErrchk(cudaPeekAtLastError());
		gpuErrchk(cudaDeviceSynchronize());
		err = cudaGetLastError();
		if (cudaSuccess != err) {
			cout << "GPU memory allocation: " << cudaGetErrorString(err)
					<< endl;
			exit(-1);
		}
	}

	gettimeofday(&endAccesstime, NULL);
	double totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Reconst on " << GPUs << " GPUs";
	if (usingCPU)
		cout << " and " << numOfThreadsCPU << " CPU  threads";
	cout << ": data transfer HtoD time  = " << totTime << " seconds." << endl;

	gettimeofday(&startAccesstime, NULL);
	omp_set_num_threads(GPUs);
#pragma omp parallel  num_threads(GPUs)
//	for (int gpuID = 0; gpuID < GPUs; gpuID++)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);

		Multi_GPU_CPU_QuaternionReconstructionKernel<<<noOfBlocks, noOfThreads>>>(
				d_r[gpuID], d_Theta1[gpuID], d_SR[gpuID], d_SB[gpuID],
				d_SG[gpuID], d_H1R[gpuID], d_H1B[gpuID], d_H1G[gpuID],
				d_h1[gpuID], d_h2[gpuID], d_h3[gpuID], shares[gpuID].i,
				shares[gpuID].j, MAX, SIZE, numOfThreads);
	}

	struct timeval startAccesstime1, endAccesstime1;
	gettimeofday(&startAccesstime1, NULL);

	size_t a_size = sizeof(double) * size_t(SIZE * SIZE);

	h_1 = (double*) malloc(a_size);
	h_2 = (double*) malloc(a_size);
	h_3 = (double*) malloc(a_size);
	for (int i = 0; i < SIZE * SIZE; i++)
		h_1[i] = h_2[i] = h_3[i] = 0;

	if (usingCPU) {
		omp_set_num_threads(numOfThreadsCPU);
#pragma omp parallel private(sum1, sum2,sum3,s1,s2,s3)  num_threads(numOfThreadsCPU)
//	for (int threadID = 0; threadID < numOfThreadsCPU; threadID++)
		{
			int threadID = omp_get_thread_num();
			int totNumofIterations = shares[shares.size() - 1].j
					- shares[shares.size() - 1].i + 1;

			int size = floor(1.0 * totNumofIterations / numOfThreadsCPU);
			if (size == 0)
				size = 1;

			unsigned localStart = shares[shares.size() - 1].i + size * threadID;
			unsigned localEnd = localStart + size - 1;

			if (threadID == numOfThreadsCPU - 1
					&& localEnd < shares[shares.size() - 1].j)
				localEnd = shares[shares.size() - 1].j;
			if (localEnd > shares[shares.size() - 1].j)
				localEnd = shares[shares.size() - 1].j;

			double PI = 3.1415926535897;

			int pmax, qmax;
			thrust::complex<double> ii(0, 1);
			thrust::complex<double> R1, R2;

			pmax = qmax = MAX;
			int i, j, temp_index;
			int p_index = 0;
			int q_index = 0;

			for (int it = localStart; it <= localEnd; it++) {
				i = it / SIZE + 1;
				j = it % SIZE + 1;

				for (int p = 1; p <= pmax; p++) {

					R1 = exp(p * 2.0 * ii * PI * pow(r[(i - 1)][(j - 1)], 2.0));
					R2 = exp(p * Theta1[(i - 1)][(j - 1)] * ii);

					p_index = p * (MAX + 1);
					s1 = s1 + (h_SR[p_index]) * (R1)
							+ thrust::conj(h_SR[p_index]) * conj(R1)
							+ (h_SR[p]) * (R2) + conj(h_SR[p] * (R2));
					s2 = s2 + (h_SG[p_index]) * (R1)
							+ conj(h_SG[p_index]) * conj(R1) + (h_SG[p]) * (R2)
							+ conj(h_SG[p] * (R2));
					s3 = s3 + (h_SB[p_index]) * (R1)
							+ conj(h_SB[p_index]) * conj(R1) + (h_SB[p]) * (R2)
							+ conj(h_SB[p] * (R2));

					for (int q = 1; q <= qmax; q++) {

						R2 = exp(q * Theta1[i - 1][j - 1] * ii);

						q_index = p * (MAX + 1) + q;
						sum1 = sum1 + h_SR[q_index] * R1 * R2
								+ conj(h_SR[q_index] * R1 * R2)
								+ h_H1R[q_index] * conj(R1) * (R2)
								+ conj(h_H1R[q_index]) * R1 * conj(R2);
						sum2 = sum2 + h_SG[q_index] * R1 * R2
								+ conj(h_SG[q_index] * R1 * R2)
								+ h_H1G[q_index] * conj(R1) * (R2)
								+ conj(h_H1G[q_index]) * R1 * conj(R2);
						sum3 = sum3 + h_SB[q_index] * R1 * R2
								+ conj(h_SB[q_index] * R1 * R2)
								+ h_H1B[q_index] * conj(R1) * (R2)
								+ conj(h_H1B[q_index]) * R1 * conj(R2);
					}
				}
				temp_index = (i - 1) * SIZE + (j - 1);

				h_3[temp_index] = (sum3 + s3 + h_SB[0]).real();
				h_2[temp_index] = (sum2 + s2 + h_SG[0]).real();
				h_1[temp_index] = (sum1 + s1 + h_SR[0]).real();

				sum3 = 0.0;
				sum2 = 0.0;
				sum1 = 0.0;
				s1 = s2 = s3 = 0;
			}
		}

		gettimeofday(&endAccesstime1, NULL);

		totTime = (double) ((endAccesstime1.tv_usec - startAccesstime1.tv_usec)
				/ 1.0e6 + endAccesstime1.tv_sec - startAccesstime1.tv_sec);
		cout << "partial cpu time = " << totTime << endl;
	}
	omp_set_num_threads(GPUs);
#pragma omp parallel  num_threads(GPUs)
	//for (int gpuID = 0; gpuID < GPUs; gpuID++)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);

		gpuErrchk(cudaPeekAtLastError());
		gpuErrchk(cudaDeviceSynchronize());
		err = cudaGetLastError();
		if (cudaSuccess != err) {
			cout << "Reconst kernel: " << cudaGetErrorString(err) << endl;
			exit(-1);
		}
	}

	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image reconst. on " << GPUs << " GPU(s) running time  = "
			<< totTime << " seconds." << " The number of threads per GPU : "
			<< totNoOfThreads << endl;

	gettimeofday(&startAccesstime, NULL);

	a_size = sizeof(double) * size_t(SIZE * SIZE);
	double* h_h1 = (double*) malloc(a_size);
	double* h_h2 = (double*) malloc(a_size);
	double* h_h3 = (double*) malloc(a_size);

	for (int gpuID = 0; gpuID < GPUs; gpuID++) {
		cudaSetDevice(gpuID);

		cudaMemcpy(h_h1, d_h1[gpuID], a_size, cudaMemcpyDeviceToHost);
		cudaMemcpy(h_h2, d_h2[gpuID], a_size, cudaMemcpyDeviceToHost);
		cudaMemcpy(h_h3, d_h3[gpuID], a_size, cudaMemcpyDeviceToHost);
		for (int i = 0, k = 0; i < SIZE; i++) {
			for (int j = 0; j < SIZE; j++) {

				h_1[k] += h_h1[k];
				h_2[k] += h_h2[k];
				h_3[k] += h_h3[k];
				k++;
			}
		}
		gpuErrchk(cudaPeekAtLastError());
		gpuErrchk(cudaDeviceSynchronize());
		err = cudaGetLastError();
	}
	gettimeofday(&endAccesstime, NULL);
	totTime = (double) ((endAccesstime.tv_usec - startAccesstime.tv_usec)
			/ 1.0e6 + endAccesstime.tv_sec - startAccesstime.tv_sec);
	cout << "Image Reconst on GPU: data transfer DtoH time  = " << totTime
			<< " seconds." << endl;

	omp_set_num_threads(GPUs);
#pragma omp parallel  num_threads(GPUs)
	//for (int gpuID = 0; gpuID < GPUs; gpuID++)
	{
		int gpuID = omp_get_thread_num();
		cudaSetDevice(gpuID);
		cudaFree(d_SB[gpuID]);
		cudaFree(d_SG[gpuID]);
		cudaFree(d_SR[gpuID]);
		cudaFree(d_H1B[gpuID]);
		cudaFree(d_H1G[gpuID]);
		cudaFree(d_H1R[gpuID]);
	}
}

int main(int argc, const char **argv) {
	if (!set_inputs(argc, argv))
		return 0;

	totNoOfThreads = noOfBlocks * noOfThreads;
	omp_set_dynamic(0);
	omp_set_num_threads(numOfThreadsCPU);

#pragma omp parallel
	{
		int ID = omp_get_thread_num();
	}

	//image independent tasks
	computeIpIq();
	transferImageIndependentDataToGPUs();

	//image dependent tasks
	Quaternion_moment_computation_MultiGPU_CPU();

	if (numOfGPUs == 1 && !usingCPU)
		GPU_Image_Reconstruct_Computation();
	else {
		GPU_CPU_Reconstruct(1, 100, 19);
	}
	//evaluating the moment accuracy numerically and visually
	errorComputing();
	saveReconstructedImage();
	cout << "All tasks are done!" << endl;

	return 0;
}
